#!/bin/sh
# -*- tcl -*-
# The next line is executed by /bin/sh, but not tcl \
exec tclsh "$0" ${1+"$@"}

array set defaultParams {}
array set imageParams   {}


# ParamName:Name BitOffs BitSize DefVal ImageVal
array set paramsDB      {}
array set imageCrspace  {}
array set crSections    {}

array set groups {}

proc buildParams {fwFile} {
    global paramsDB
    global groups
    global env
    
    set env(MIC_PAR_DEBUG) 1
    set emptyIni "/tmp/empty.[pid].ini"
    set parout   "/tmp/parout"
    set mlxburn [exec which mlxburn]
    exec -- echo "" > $emptyIni
    if {[catch {exec -- tclsh $mlxburn -nofs_img -fw $fwFile -c $emptyIni -wrimage /dev/null > $parout} e]} {
	puts "-E- mlxburn activation failed. May have some details in file $parout"
	exit 1
    }
    set fh ""
    
    if {[catch "set fh [open $parout]" e]} {
        puts "-E- Failed to open $parout: $e"
        exit 1
    }
    
    while { ![eof $fh] } {
	gets $fh line
        # Par:Diagnostic.LinkDownedCntInfoTh                ; CrSpace 0x03f33c.16 /size:08/ <= 0x      ff. boot2=0
        if {[regexp {^Par:} $line]} {
            if {[regexp {^Par:([^\.]+)\.(\S+)\s*; CrSpace 0x([0-9a-f]+)\.(\d+) /size:(\d+)/ <= 0x *([0-9a-f]+)} \
                 $line d1 grp par crAddr bitOffs bitSize val]} {
                set crAddr  [format "0x%08x" [expr 0x$crAddr]]
                set val     "0x$val"
                set bitSize [string trimleft $bitSize 0]
                set bitOffs [string trimleft $bitOffs 0]
		if {$bitOffs == ""} {set bitOffs 0}
                
		set grp [string trim $grp]
                set groups($grp) 1
                set parName "$grp.$par"
                set paramsDB($parName:Name) $parName
                set paramsDB($parName:CrLoc) [list $crAddr $bitOffs $bitSize]
                set paramsDB($parName:DefVal) $val
		if {![regexp {Enum:(\S*)$} $line d1 enum]} {
		    puts "-E- Enum declaration not found on param info line. Probably an old MFT version - please update"
		    exit 1
		}
		set paramsDB($parName:Enum) $enum
            } else {
                puts "-E- Par line does not match par regexp: $line"
                exit 1
            }
        }
    }

    close $fh
    
    file delete -- $emptyIni
    # $parout
}

proc getCrSections {imgFile} {
    global crSections
    
    set fh 0
    if {[catch "set fh [open $imgFile]" e]} {
        puts "-E- Failed to open $imgFile: $e"
        exit 1
    }
    
    set inSection 0
    set sectType ""
    while { ![eof $fh] } {
	gets $fh line
	
	if {$inSection} {
	    if {[regexp {^END } $line]} {
		# pop the lase crc word
		set sectDw [lrange $sectDw 0 end-1]
		set inSection 0
		lappend crSections($sectType) $sectDw
	    } else {
		if {![regexp {^\s*\S\S \S\S \S\S \S\S\s*$} $line]} {
		    puts "-E- bad data line format in $imgFile: \"$line\". Expecting 4 bytes."
		    exit 1
		}
		set dw [join $line ""]
		set dw "0x$dw"
		lappend sectDw $dw
	    }
	} elseif {[regexp {^START .*(NBI|HBI|RMW|IWI).+} $line d1 sectType]} {
	    set inSection 1
	    # eat the header rows
	    gets $fh line
	    gets $fh line
	    
	    set sectDw {}
	}
    }
    
    close $fh
}


proc buildCrspaceFromImg {imgFile} {
    global imageCrspace
    global crSections
    
    getCrSections $imgFile
    
    if {[info exist crSections(HBI)]} {
	foreach dwl $crSections(HBI) {
	    if {[llength $dwl] != 3} {
	        puts "-E- Bad image file format: HBI section data is not 3 DWs length."
	        exit 1
	    }
	    set baseAddr [format {0x%08x} [expr {[lindex $dwl 0]}]]
	    set len      [expr {[lindex $dwl 1]}]
	    set data     [lindex $dwl 2]
	    for {set i 0} {$i < $len} {incr i 4} {
		set addr [format {0x%08x} [expr {$baseAddr + $i}]]
		set imageCrspace($addr) $data
	    }
	}
    }
    
     if {[info exist crSections(NBI)]} {
	foreach dwl $crSections(NBI) {
	    set addr [format {0x%08x} [expr {[lindex $dwl 0]}]]
	    foreach data [lrange $dwl 1 end] {
		set imageCrspace($addr) $data
		set addr [format {0x%08x} [expr {$addr + 4}]]
	    }
	}
    }   
    
    if {[info exist crSections(RMW)]} {
	foreach dwl $crSections(RMW) {
	    if {[llength $dwl] % 3} {
		puts "-E- Bad image file format: RMW section data is not 3 DWs aligned."
		exit 1
	    }
	    
	    foreach {addr data mask} $dwl {
		if {[info exist imageCrspace($addr)]} {
		    set imageCrspace($addr) [format {0x%08x} [expr {( $imageCrspace($addr) & ~$mask) | ($data & $mask)}]]
		} else {
		    set imageCrspace($addr) $data
		}
	    }
	    
	}
    }
    
    if {[info exist crSections(IWI)]} {
	foreach dwl $crSections(IWI) {
	    if {[llength $dwl] % 2} {
		puts "-E- Bad image file format: IWI section data is not 2 DWs aligned."
		exit 1
	    }
	    
	    foreach {addr data} $dwl {
		set imageCrspace($addr) $data
	    }
	}
    }
}

proc buildCrspaceFromBin {binFile} {
    global imageCrspace
    
}

proc getImageVal {crLoc} {
    global imageCrspace
    set addr [lindex $crLoc 0]
    set offs [lindex $crLoc 1]
    set size [lindex $crLoc 2]
    
    if {![info exist imageCrspace($addr)]} {
        return "*N/A*"
    }
    
    if {$size == 32} {
	set data $imageCrspace($addr)
    } else {
	set data [format 0x%x [expr {( $imageCrspace($addr) >> $offs) & ((1<<$size)-1)}]]
    }
    return $data
}

proc setImageParams {} {
    global paramsDB
    
    foreach pn_ [array names paramsDB *:Name] {
        set pn $paramsDB($pn_)
	set val [getImageVal $paramsDB($pn:CrLoc)]
	set paramsDB($pn:ImageVal) $val
	if {$paramsDB($pn:Enum) != ""} {
	    set ev ""
	    foreach n_v [split $paramsDB($pn:Enum) ","] {
		set enumVal  [lindex [split $n_v =] 1]
		set enumName [lindex [split $n_v =] 0]
		
		if {$val == $enumVal} {
		    set ev $enumName
		}
	    }
	    if {$ev != ""} {
		set  paramsDB($pn:ImageValEnum) $ev		
	    } else {
		set  paramsDB($pn:ImageValEnum) "UNKNOWN_ENUM($val)"
	    }
	}
    }
}

proc printIni {iniFile} {
    global paramsDB
    global groups
    global argv
    global argv0
    
    set iniFh ""
    
    if {$iniFile == ""} {
	set iniFh stdout
    } else {
	if {[catch "set iniFh [open $iniFile w]" e]} {
	    puts "-E- Failed to open file $iniFile: $e"
	    exit 1
	}
    }

    # header
    puts $iniFh "
;;
;; THIS INI FILE WAS AUTO-GENERATED ON [clock format [clock seconds]]
;; GENERATION COMMAND LINE: $argv0 $argv
;;
;; NOTE:
;; THIS FILE IS *NOT* 100% IDENTICAL TO THE IMAGE FILE USED TO GENERATE
;; THE ORIGINAL IMAGE. DIFFERENCES:
;; - BOOT RECORD CHANGES ARE NOT EXRTRACTED
;; - HIDDEN PARAMETERS, WHICH ARE DERIVED FROM THE INI PARAMETERS IN
;;   REGULAR IMAGE GENERATION AND ARE NOT ACCESSIBLE TO THE USER, ARE
;;   SHOWN AS IS. THE ORIGINAL INI PARAMETERS MAY NOT BE DISPLAYED.
;;
;; THE USER SHOULD PAY ATTENTION TO THE ABOVE DIFFERENCES WHEN USING THIS FILE
;; IT IS RECOMMENDED TO MANUALLY ADD \[PLL\] \[EEPROM\] \[PSID\] SECTIONS TO
;; THIS FILE IN ORDER TO GENERATE AN IDENTICAL IMAGE TO THE ORIGINAL
;;
    "
    
    foreach g [array names groups] {
        puts $iniFh "\[$g\]"
        foreach pn_ [array names paramsDB $g.*:Name] {
            set pn $paramsDB($pn_)
            set param [lindex [split $pn .] 1]
            if {$paramsDB($pn:ImageVal) != $paramsDB($pn:DefVal)} {
		if {[info exist paramsDB($pn:ImageValEnum)]} {
		    puts $iniFh "$param = $paramsDB($pn:ImageValEnum)"
		} else {
		    puts $iniFh "$param = $paramsDB($pn:ImageVal)"
		}
            }
        }
        puts $iniFh ""
    }
    
    if {$iniFh != "stdout"} {
	close $iniFh	
    }
}

#
# MAIN
#

set iniFile ""

if {[llength $argv] != 3} {
    puts "Usage: $argv0 <fw-file> <image-file> <output-ini>"
    puts "
    
  Note:
  This tool is Mellanox Internal.
  The ini file generated by this tool needs some manual patching, and should
  not be used for production.
    
  Prerequisits:
  Find the device FW version. The FW file used by this tool must be
  of the same FW version. Otherwise, results may be undefined.
	
  Example use case:

  # Read mysterious image from the device
  > spark -d /dev/mst/mtusb1 ri /tmp/device_fw.img
  
  # run the tool:
  > $argv0 ./IS3FW.BIN /tmp/device_fw.img /tmp/generated.ini
  
  # Edit the generated ini - Manually add EEPROM, PSID and PLL sections
  
  # You can now use this ini to generate fw image from a newer version.
  "
     
    
    
    exit 1
}

set fwFile  [lindex $argv 0]
set imgFile [lindex $argv 1]
set iniFile [lindex $argv 2]

set imgExt [file extension $imgFile] 

if {$imgExt == ".img"} {
    buildCrspaceFromImg $imgFile
} elseif {$imgExt == ".bin"} {
    #buildCrspaceFromBin $imgFile
    puts "-E- Unsupported image file format ($imgExt)"
    exit 1

} else {
    puts "-E- Unsupported image file format ($imgExt)"
    exit 1
}

buildParams $fwFile
setImageParams
printIni $iniFile
#parray imageCrspace
#parray paramsDB
