
source [file join [file dirname [info script]] mlx_i2c.tcl]
global DZ

proc Set_RegTable {} {
    global RegTable DZ
    catch {unset RegTable}
    array set RegTable {
        AC_Pfail  "0 none AC Pfail"
        PS_Status "0 none PS Status"
        P_Good    "0 none P Good"
        Vin_Good  "0 none Vin Good"
        OVP_12V   "0 none OVP 12V"
        UVP_12V   "0 none UVP 12V"
        Fan_Fault "0 none Fan Fault"
        OCP_12V   "0 none OCP 12V"
    }
}

proc List {_slave} {
    global RegTable DZ
    foreach name [array names RegTable] {
        Inform "-I-dump:data" -data "$name"
    }
}

proc Query {_slave {dump 0}} {
    global RegTable DZ
    set res ""
    foreach name [array names RegTable] {
        set lvl [lindex $RegTable($name) 0]
        if {$lvl} {
            continue
        }
        set address [lindex $RegTable($name) 1]
        set desc [lrange $RegTable($name) 2 end]
        set val [Read $address $_slave $name]
        lappend res "$name,$val"
        if {$dump} {
            Inform "-I-dump:data" -data "[SeperateBySpaces "$name=$val [list $desc]" 20]"
        }
    }
    return $res
}

proc Read_Handle {_list_read_address _i2c_address} {
    global RegTable DZ
    if {$_list_read_address == "all"} {
        foreach name [array names RegTable] {
            set lvl [lindex $RegTable($name) 0]
            set address [lindex $RegTable($name) 1]
            set desc [lrange $RegTable($name) 2 end]
            set val [Read $address $_i2c_address $name]
            Inform "-I-dump:data" -data "$name = $val"
        }
        return 0
    }
    foreach reg [split $_list_read_address ,] {
        if {[string is integer $reg]} {
            set name [Get_NameByReg $reg]
            set address $reg
            set val [Read $address $_i2c_address $name]
            if {$val != -1} {
                Inform "-I-dump:data" -data "$name = $val"
            }
        } elseif {[regexp {^([^-]+)-([^-]+)$} $reg . start end]} {
            if {[string is integer $start] && [string is integer $end]} {
                set current $start
                for {set i 0} {$i <= [Abs "$end - $start"]} {incr i} {
                    set name [Get_NameByReg $current]
                    set address $current
                    set val [Read $address $_i2c_address $name]
                    if {$val != -1} {
                        Inform "-I-dump:data" -data "$name = $val"
                    }
                    incr current [expr (2 * ($start < $end) ) - 1]
                }
            } else {
                Inform "-E-argv:invalid.read.address" -value $reg
            }
        } elseif {[Get_RegByName $reg] != -1 } {
            set name $reg
            set address [Get_RegByName $reg]
            set val [Read $address $_i2c_address $name]
            if {$val != -1} {
                Inform "-I-dump:data" -data "$name = $val"
            }
        } else {
            Inform "-E-argv:invalid.read.address" -value $reg
        }

    }
}

proc Write_Handle {_list_write_address_vals _i2c_address} {
    foreach pair [split $_list_write_address_vals ,] {
        if {[regexp {^([^=]+)=([0-9a-f0-F]+)$} $pair . reg val]} {
            if {[string is integer $reg]} {
                set name [Get_NameByReg $reg]
                set address $reg
                Write $address $val $_i2c_address $name 
            } elseif {[Get_RegByName $reg] != -1 } {
                set name $reg
                set address [Get_RegByName $reg]
                Write $address $val $_i2c_address $name 
            } else {
                Inform "-E-argv:invalid.write.param" -value $pair
            }
        } else {
            Inform "-E-argv:invalid.write.param" -value $pair
        }
    }
}


proc Read {_address _i2c_address _name} {
    global DZ

    if {$_address == "none"} {
        set args_list "-dev $DZ(argv:mst.device) -i2c_address $_i2c_address -address 0x7d"
        set res_low [I2C_Read $args_list]
        set args_list "-dev $DZ(argv:mst.device) -i2c_address $_i2c_address -address 0x7f"
        set res_high [I2C_Read $args_list]

        set res ${res_high}${res_low}

        array set SignalMap {
            AC_Pfail  7
            PS_Status 6
            P_Good    5
            Vin_Good  4
            OVP_12V   3
            UVP_12V   2
            Fan_Fault 1
            OCP_12V   0
        }

        switch $_name {
                AC_Pfail  -
                PS_Status -
                P_Good    -
                Vin_Good  -
                OVP_12V   -
                UVP_12V   -
                Fan_Fault -
                OCP_12V  {
                    if {$res == -1} {
                        return -1
                    }
                    set res_bin [DupChar 0 8][Hex2Bin $res]
                    return [lindex $res_bin end-$SignalMap($_name)]
                }
                default {
                    return -1
                }
            }
        }
    }

    set args_list "-dev $DZ(argv:mst.device) -i2c_address $_i2c_address -address $_address"
    return [I2C_Read $args_list]
}

proc Write {_address _val _i2c_address _name} {
    global DZ

    if {$_address == "none"} {
        switch $_name {
            default {
                return -1
            }
        }
    }

    set args_list "-dev $DZ(argv:mst.device) -i2c_address $_i2c_address -address $_address -data $_val"
    return [I2C_Write $args_list]
}

proc Get_NameByReg {_address} {
    global RegTable
    foreach name [array names RegTable] {
        set val [lindex $RegTable($name) 1]
        if {$val == $_address} {
            return $name
        }
    }
    return "<No Name>"
}

proc Get_RegByName {_name} {
    global RegTable
    if {![info exists RegTable($_name)]} {
        puts "NONE VALID NAME $_name"
        return -1
    }
    return [lindex $RegTable($_name) 1]
}

Set_RegTable

if {[info exists DZ(var:tool.name)]} {
    return 0
}

### Pre runs
Initialize_MlxI2C
Start_MlxI2C

set mode ""
if {[info exists DZ(argv:read.address)]} {
    set mode _Read
}
if {[info exists DZ(argv:write.address.val)]} {
    set mode _Write
}
if {[info exists DZ(argv:list)]} {
    set mode _List
}
if {[info exists DZ(argv:query)]} {
    set mode _Query
}

Inform "-I-sharq:Mode.header" -mode $mode

switch $mode {
    "_List" {
        List $DZ(argv:slave.address)
    }
    "_Query" {
        Query $DZ(argv:slave.address) 1
    }
    "_Read" {
        Read_Handle $DZ(argv:read.address) $DZ(argv:slave.address)
    }
    "_Write" {
        Write_Handle $DZ(argv:write.address.val) $DZ(argv:slave.address)
    }
}

Finish_MlxI2C
