source [file join [file dirname [info script]] mlx_i2c.tcl]
global DZ

proc Set_RegTable {} {
    global RegTable DZ
    catch {unset RegTable}
    array set RegTable {
        IP0  "0 0x00 read only Input Port register bank 0"
        IP1  "0 0x01 read only Input Port register bank 1"
        IP2  "0 0x02 read only Input Port register bank 2"
        IP3  "0 0x03 read only Input Port register bank 3"
        IP4  "0 0x04 read only Input Port register bank 4"
        OP0  "0 0x08 read/write Output Port register bank 0"
        OP1  "0 0x09 read/write Output Port register bank 1"
        OP2  "0 0x0A read/write Output Port register bank 2"
        OP3  "0 0x0B read/write Output Port register bank 3"
        OP4  "0 0x0C read/write Output Port register bank 4"
        PI0  "0 0x10 read/write Polarity Inversion register bank 0"
        PI1  "0 0x11 read/write Polarity Inversion register bank 1"
        PI2  "0 0x12 read/write Polarity Inversion register bank 2"
        PI3  "0 0x13 read/write Polarity Inversion register bank 3"
        PI4  "0 0x14 read/write Polarity Inversion register bank 4"
        IOC0 "0 0x18 read/write I/O Configuration register bank 0"
        IOC1 "0 0x19 read/write I/O Configuration register bank 1"
        IOC2 "0 0x1A read/write I/O Configuration register bank 2"
        IOC3 "0 0x1B read/write I/O Configuration register bank 3"
        IOC4 "0 0x1C read/write I/O Configuration register bank 4"
        MSK0 "0 0x20 read/write Mask Interrupt register bank 0"
        MSK1 "0 0x21 read/write Mask Interrupt register bank 1"
        MSK2 "0 0x22 read/write Mask Interrupt register bank 2"
        MSK3 "0 0x23 read/write Mask Interrupt register bank 3"
        MSK4 "0 0x24 read/write Mask Interrupt register bank 4"


        PIN0  "1 none pin0"
        PIN1  "1 none pin1"
        PIN2  "1 none pin2"
        PIN3  "1 none pin3"
        PIN4  "1 none pin4"
        PIN5  "1 none pin5"
        PIN6  "1 none pin6"
        PIN7  "1 none pin7"
        PIN8  "1 none pin8"
        PIN9  "1 none pin9"
        PIN10 "1 none pin10"
        PIN11 "1 none pin11"
        PIN12 "1 none pin12"
        PIN13 "1 none pin13"
        PIN14 "1 none pin14"
        PIN15 "1 none pin15"
        PIN16 "1 none pin16"
        PIN17 "1 none pin17"
        PIN18 "1 none pin18"
        PIN19 "1 none pin19"
        PIN20 "1 none pin20"
        PIN21 "1 none pin21"
        PIN22 "1 none pin22"
        PIN23 "1 none pin23"
        PIN24 "1 none pin24"
        PIN25 "1 none pin25"
        PIN26 "1 none pin26"
        PIN27 "1 none pin27"
        PIN28 "1 none pin28"
        PIN29 "1 none pin29"
        PIN30 "1 none pin30"
        PIN31 "1 none pin31"
        PIN32 "1 none pin32"
        PIN33 "1 none pin33"
        PIN34 "1 none pin34"
        PIN35 "1 none pin35"
        PIN36 "1 none pin36"
        PIN37 "1 none pin37"
        PIN38 "1 none pin38"
        PIN39 "1 none pin39"
    }
}

proc List {_slave} {
    global RegTable DZ
    foreach name [lsort -command NameSort [array names RegTable]] {
        if {[regexp {^PIN.*$} $name]} {
            continue
        }
        Inform "-I-dump:data" -data "$name"
    }
}

proc Query {_slave {dump 0}} {
    global RegTable DZ
    set res ""
    foreach name [lsort -command NameSort [array names RegTable]] {
        set lvl [lindex $RegTable($name) 0]
        if {$lvl} {
            continue
        }
        set address [lindex $RegTable($name) 1]
        set desc [lrange $RegTable($name) 2 end]
        set val [Read $address $_slave $name]
        lappend res "$name,$val"
        if {$dump} {
            Inform "-I-dump:data" -data "[SeperateBySpaces "$name=$val [list $desc]" 20]"
        }
    }
    return $res
}

proc Read_Handle {_list_read_address _i2c_address {silent 0}} {
    global RegTable DZ
    if {$_list_read_address == "all"} {
        foreach name [lsort -command NameSort [array names RegTable]] {
            set lvl [lindex $RegTable($name) 0]
            set address [lindex $RegTable($name) 1]
            set desc [lrange $RegTable($name) 2 end]
            set val [Read $address $_i2c_address $name]
            Inform "-I-dump:data" -data "$name = $val"
        }
        return 0
    }
    foreach reg [split $_list_read_address ,] {
        if {[string is integer $reg]} {
            set name [Get_NameByReg $reg]
            set address $reg
            set val [Read $address $_i2c_address $name]
            if {$val != -1} {
                Inform "-I-dump:data" -data "$name = $val"
            }
        } elseif {[regexp {^([^-]+)-([^-]+)$} $reg . start end]} {
            if {[string is integer $start] && [string is integer $end]} {
                set current $start
                for {set i 0} {$i <= [Abs "$end - $start"]} {incr i} {
                    set name [Get_NameByReg $current]
                    set address $current
                    set val [Read $address $_i2c_address $name]
                    if {$val != -1} {
                        Inform "-I-dump:data" -data "$name = $val"
                    }
                    incr current [expr (2 * ($start < $end) ) - 1]
                }
            } else {
                Inform "-E-argv:invalid.read.address" -value $reg
            }
        } elseif {[Get_RegByName $reg] != -1 } {
            set name $reg
            set address [Get_RegByName $reg]
            set val [Read $address $_i2c_address $name]
            if {$silent} { return $val }
            if {$val != -1} {
                Inform "-I-dump:data" -data "$name = $val"
            }
        } else {
            Inform "-E-argv:invalid.read.address" -value $reg
        }
    }
}

proc Write_Handle {_list_write_address_vals _i2c_address} {
    foreach pair [split $_list_write_address_vals ,] {
        if {[regexp {^([^=]+)=([0-9a-f0-F]+)$} $pair . reg val]} {
            if {[string is integer $reg]} {
                set name [Get_NameByReg $reg]
                set address $reg
                Write $address $val $_i2c_address $name 
            } elseif {[Get_RegByName $reg] != -1 } {
                set name $reg
                set address [Get_RegByName $reg]
                Write $address $val $_i2c_address $name 
            } else {
                Inform "-E-argv:invalid.write.param" -value $pair
            }
        } else {
            Inform "-E-argv:invalid.write.param" -value $pair
        }
    }
}


proc Read {_address _i2c_address _name} {
    global DZ
    if {$_address == "none"} {
        if {![regexp {^PIN([0-9]+)$} $_name . pinNum ]} {
            return -1
        }
        set status [Read [expr 0x18 + ($pinNum / 8)] $_i2c_address "TMP"]
        if {$status == -1} {
            return -1
        }
        set status_bin [DupChar 0 8][Hex2Bin $status]
        set pin_status [string index $status_bin end-[expr $pinNum % 8]]
        if {$pin_status} {
            set status_1 [Read [expr 0x0 + ($pinNum / 8)] $_i2c_address "TMP"]
            set status_bin_1 [DupChar 0 8][Hex2Bin $status_1]
            set pin_status_1 [string index $status_bin_1 end-[expr $pinNum % 8]]
            set status "INPUT"
        } else {
            set status_1 [Read [expr 0x8 + ($pinNum / 8)] $_i2c_address "TMP"]
            set status_bin_1 [DupChar 0 8][Hex2Bin $status_1]
            set pin_status_1 [string index $status_bin_1 end-[expr $pinNum % 8]]
            set status "OUTPUT"
        }
        return "$status $pin_status_1"
    }
    set args_list "-dev $DZ(argv:mst.device) -i2c_address $_i2c_address -address $_address"
    return [I2C_Read $args_list]
}

proc Write {_address _val _i2c_address _name} {
    global DZ
    if {$_address == "none"} {
        if {![regexp {^PIN([0-9]+)$} $_name . pinNum ]} {
            return -1
        }
        set status [Read [expr 0x18 + ($pinNum / 8)] $_i2c_address "TMP"]
        if {$status == -1} {return -1}
        set status_bin [DupChar 0 8][Hex2Bin $status]
        set pin_status [string index $status_bin end-[expr $pinNum % 8]]
        if {$pin_status} {
            Inform "-W-gpio:writing.to.input.pin" -name $_name -val $_val
        }
        if {($_val != 0) &&($_val != 1)} {
            Inform "-E-gpio:writing.illegal.pin.val" -val $_val
        }
        set output_val [Read [expr 0x8 + ($pinNum / 8)] $_i2c_address "TMP"]
        if {$output_val == -1} {return -1}
        set output_val [Bin2Hex [string replace [Hex2Bin $output_val] end-[expr $pinNum % 8] end-[expr $pinNum % 8] $_val]]
        set args_list "-dev $DZ(argv:mst.device) -i2c_address $_i2c_address -address [expr 0x8 + ($pinNum / 8)] -data $output_val"
    } else {
        set args_list "-dev $DZ(argv:mst.device) -i2c_address $_i2c_address -address $_address -data $_val"
    }
    return [I2C_Write $args_list]

    global DZ
    set args_list "-dev $DZ(argv:mst.device) -i2c_address $_i2c_address -address $_address -data $_val"
    return [I2C_Write $args_list]
}

proc Get_NameByReg {_address} {
    global RegTable
    foreach name [array names RegTable] {
        set val [lindex $RegTable($name) 1]
        if {$val == $_address} {
            return $name
        }
    }
    return "<No Name>"
}

proc Get_RegByName {_name} {
    global RegTable
    if {![info exists RegTable($_name)]} {
        puts "NONE VALID NAME $_name"
        return -1
    }
    return [lindex $RegTable($_name) 1]
}

proc NameSort {_a _b} {
    set value_prename_list "IP OP PI IOC MSK PIN"
    regexp {^([^0-9]+)(.*)$} $_a . prename_a indx_a
    regexp {^([^0-9]+)(.*)$} $_b . prename_b indx_b

    set value_prename_a [lsearch $value_prename_list $prename_a]
    set value_prename_b [lsearch $value_prename_list $prename_b]

    if {$value_prename_a < $value_prename_b} {
        return -1
    } elseif {$value_prename_a > $value_prename_b} {
        return 1
    } else {
        if {$indx_a < $indx_b} {
            return -1
        } else {
            return 1
        }
    }
}

Set_RegTable

if {[info exists DZ(var:tool.name)]} {
    return 0
}

### Pre runs
Initialize_MlxI2C
Start_MlxI2C

set mode ""
if {[info exists DZ(argv:read.address)]} {
    set mode _Read
}
if {[info exists DZ(argv:write.address.val)]} {
    set mode _Write
}
if {[info exists DZ(argv:list)]} {
    set mode _List
}
if {[info exists DZ(argv:query)]} {
    set mode _Query
}

Inform "-I-sharq:Mode.header" -mode $mode
switch $mode {
    "_List" {
        List $DZ(argv:slave.address)
    }
    "_Query" {
        Query $DZ(argv:slave.address) 1
    }
    "_Read" {
        Read_Handle $DZ(argv:read.address) $DZ(argv:slave.address)
    }
    "_Write" {
        Write_Handle $DZ(argv:write.address.val) $DZ(argv:slave.address)
    }
}

Finish_MlxI2C

