#Usage:
#         i2c [OPTIONS] <device> <cmd> <i2c_addr> <addr> [<data>]
proc I2C_Handler {args} {
    global DZ
    set args [lindex $args 0]
    set dev [Get_OffsetElement $args -dev]
    set cmd [Get_OffsetElement $args -cmd]
    set i2c_address [Get_OffsetElement $args -i2c_address]
    set address [Get_OffsetElement $args -address]
    set data [Get_OffsetElement $args -data]

    set address 0x[format %x [expr $address]]

    set cmd_line $DZ(var:cmd.i2c)
    if {$dev == ""} {
        set dev $DZ(argv:mst.device)
    }

    foreach element "-a -d -x" {
        set val [Get_OffsetElement $args $element]
        if {$val != ""} {
            lappend cmd_line $element
            lappend cmd_line $val
        }
    }

    lappend cmd_line $dev
    lappend cmd_line $cmd
    lappend cmd_line $i2c_address
    lappend cmd_line $address
    if {$cmd == "w"} {
        set data 0x[format %x [expr $data]]
        lappend cmd_line $data
    }

    Inform "-V-exec:run.command" -cmd $cmd_line

    if {[catch {set res [eval exec $cmd_line]} a]} {
        Inform "-V-exec:command.failed" -err $a
        return -1
    }
    
    Inform "-V-exec:command.res" -res $res
    return $res
}

proc I2C_Read {args} {
    set res [I2C_Handler [join "$args -cmd r"]]
    if {$res == -1} {
        return -1
    } else {
        return 0x$res
    }
}

proc I2C_Write {args} {
    return [I2C_Handler [join "$args -cmd w"]]
}

proc Get_ListOfDevices {} {
    global DZ 
    if {![file exists $DZ(var:cmd.mdevices)]} {
        Inform "-E-mst:can.not.retrive.mdevices"
        return ""
    }
    set res ""
    if {[catch {set tmpRes [exec $DZ(var:cmd.mdevices)]} e]} {
        Inform "-E-mst:mdevices.list.empty"
    }
    for {set i 0} {$i < [llength $tmpRes]} {incr i 4} {
        lappend res [lindex $tmpRes $i]
    }
    Inform "-V-exec:run.command" -cmd $DZ(var:cmd.mdevices)
    Inform "-V-exec:command.res" -res $tmpRes
    return $res
}

proc Get_ListOfSlaves {_mst_dev} {
    global DZ 

    if {![file exists $DZ(var:cmd.i2c_slaves)]} {
        Inform "-E-isw:can.not.retrive.i2c_slaves"
        return ""
    }

    set res ""
    if {[catch {set tmpRes [eval exec $DZ(var:cmd.i2c_slaves) -d $_mst_dev]} e]} {
        Inform "-E-isw:i2c_slaves.list.empty"
    }
    for {set i 0} {$i < [llength $tmpRes]} {incr i} {
        lappend res [lindex $tmpRes $i]
    }
    Inform "-V-exec:run.command" -cmd $DZ(var:cmd.i2c_slaves)
    Inform "-V-exec:command.res" -res $tmpRes
    return $res
}
