#!/bin/sh
# -*-Tcl-*- (for emacs) \
# if [ `id -u` -ne 0 ]; then echo "-E- Only root can run this application"; exit 1; fi; \
exec tclsh "$0" "$@"

#
# mst_ib_add :
# Parse the given lst file and print nice device names for
# mst inband access
#

# lst line format:
# { CA Ports:02 SystemGUID:0002c90001777053 NodeGUID:0002c90001777050 PortGUID:0002c90001777051 VenID:000002C9 DevID:634A0000 Rev:000000A0 {ss005 HCA-1} LID:0001 PN:01 } { SW Ports:18 SystemGUID:0000000000001230 NodeGUID:0000000000000001 PortGUID:0000000000000001 VenID:000002C9 DevID:B9240000 Rev:000001A1 {MT47396 Infiniscale-III Mellanox Technologies} LID:0003 PN:01 } PHY=4x LOG=ACT SPD=2.5

# IB_DB array
# Major key: Node guid
# Minor keys: Desc Type Dev LID

array set IB_DB {}

proc addNode {portData} {
    global IB_DB

    if {![ regexp { NodeGUID:([[:xdigit:]]+)} $portData d1 nguid]} {
	error "Bad port format - Missing NodeGUID: $portData"
    }

    # If this node already exist in db - no need to re-add it
    if {[info exist IB_DB($nguid:NodeGUID)]} {
        return
    }

    if {![ regexp { DevID:([[:xdigit:]]+)} $portData d1 devId]} {
	error "Bad link format - Missing DevID: $portData"
    }
    # dev id seem to have 4 extra zeroes on the right.
    # HACK - ignore these zeroes.
    set devId [string range $devId 0 3]
    set devId [expr "0x$devId"]

    if {![ regexp { LID:([[:xdigit:]]+)} $portData d1 lid]} {
	error "Bad link format - Missing LID: $portData"
    }
    set lid "0x$lid"

    set type [lindex $portData 0]
    set desc [lindex $portData 8]

    if {$type == "CA-SM"} {
            set type "CA"
    }

    set IB_DB($nguid:NodeGUID)  $nguid
    set IB_DB($nguid:LID)       $lid
    set IB_DB($nguid:Desc)      $desc
    set IB_DB($nguid:Type)      $type
    set IB_DB($nguid:DevId)     $devId
}

proc addMstIb {hcaId port} {
    global IB_DB

    set suffix ""
    if {$hcaId != ""} {
	append suffix ",$hcaId"
	if {$port != 0} {
	    append suffix ",$port"
	}
    }

    foreach nguid [array names IB_DB *:NodeGUID] {
        set ng [lindex [split $nguid ":"] 0]

        set dev "$IB_DB($ng:Type)_MT$IB_DB($ng:DevId)_"

        if {$IB_DB($ng:Type) == "CA"} {
            set desc $IB_DB($ng:Desc)
            regsub -all { } $desc "_" desc
            append dev $desc
            append dev "_"
        }

        append dev "lid-$IB_DB($ng:LID)"
        puts "$dev$suffix"
    }
}

if {[llength $argv] > 3 || [llength $argv] < 1 || [lindex $argv 0] == "-h"} {
    puts stderr "  Usage: $argv0 <lst_file> \[hca_id\] \[exit_port_num\]"
    exit 1
}

set hcaId ""
set portNum 0
set lstFile [lindex $argv 0]

if {[llength $argv] > 1} {
    set hcaId [lindex $argv 1]
}
if {[llength $argv] > 2} {
    set portNum [lindex $argv 2]
}

if {[catch {set fh [open $lstFile r]} e]} {
    puts stderr "-E- Failed to open file $lstFile: $e"
    exit 1
}

set notActiveLinksFound 0
set lineNum 0
while {[gets $fh sLine] >= 0} {
    incr lineNum

    if {[regexp { LOG=(\S+)} $sLine d1 portState]} {
        if {$portState != "ACT"} {
	    set notActiveLinksFound 1
	    continue
	}
    } else {
	continue
    }
    if {[catch {addNode [lindex $sLine 0] ; addNode [lindex $sLine 1]} e ]} {
        puts stderr "-E- $lstFile:$lineNum: Bad LST line format: $e"
        exit 1
    }
}

addMstIb $hcaId $portNum

close $fh

if {$notActiveLinksFound} {
    puts stderr "-W- Non active links found in the IB fabric. These links are not added to the device list. Make sure that a subnet-manager is running."
}
