/*
 *
 *  mtcr.c - Mellanox Software tools (mst) driver implementation
 *
 */

#include <stdio.h>

#include <arpa/inet.h>
#include <asm/byteorder.h>
#include <dirent.h>
#include <endian.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <signal.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/un.h>
#include <unistd.h>


#ifndef MST_CPU_ppc_8xx
#include "usbdescriptor.h"
#endif
#include "mtcr.h"
#include "mtcr_types.h"
#include "mtcr_mif.h"

#include "mtcr_ib.h"
#include "mtcr_i2c.h"
#include "mst_pci.h"
#include "mst_pciconf.h"
#include "mst_calbr.h"
#include "usbioctl.h"

#ifdef MST_CPU_ppc_8xx
 #include "mst_ppc.h"
#endif

#ifdef ENABLE_MST_DEV_I2C
 #include <linux/i2c.h>
 #include <linux/i2c-dev.h>
#endif

#define CLEAR(st) memset(&(st), 0, sizeof(st))
#define MST_DIR   "/dev/mst"
#define DEVASYS_V 0x0abf
#define DEVASYS_P 0x3e9
#define DIMAX_V   0x0abf
#define DIMAX_P   0x3370
#define VENID     "Vendor="
#define PRDID     "ProdID="
#define BUS       "Bus="
#define DEV       "Dev#="
#define STRLEN    512
#define MAX_USB_RD_RETRY 10

/*
 * We do this ugly trick here just
 * not to export names from tcp.c
 */
#define INSIDE_MTCR static
#include "tcp.c"


static void  sig_pipe(int signo);
static int get_device_flags(const char *name);


/*
 * This function is never called.
 * It needs tio shutup stupid compiler warnings
 * "defined but not used" for static functions
 */
void mst_dummy()
{
    readn(-1, 0, 0);
    readnl(-1, 0, 0);
    writenl(-1, "");
    open_serv_connection(0);
}

static int remote_open(const char *name, mfile *mf, DType dtype)
{
    int  con, port, i;
    char buf[40], *name_p, nameb[1024];
    char *dev_p, *port_p;

    /*  Skip path */
    if ((name_p = strrchr(name, '/')) == 0)
        name_p = (char *)name;
    else
        name_p++;

    /*  Copy remote name, change '@' to '/' */
    for (i=0; i<1022 && name_p[i]; i++)
        nameb[i] = (name_p[i] == '@') ? '/' : name_p[i];
    nameb[i] = '\0';

    /*  Device and port pointers */
    dev_p = strchr(nameb, ',');
    port_p = strchr(nameb, ':');
    if (!dev_p || !port_p)
    {
        errno = EINVAL;
        return -1;
    }
    *port_p = '\0';
    port_p++;
    dev_p++;

    /*  Non-verbose */
    logset(0);

    /*  Open connection */
    port = strtol(port_p, 0, 10);
    con = open_cli_connection(nameb, port);
    if (con < 0)
        return -1;

    /*  Open device */

    sprintf(buf, "O 0x%08x ", dtype);
    if (writen(con, buf, 13) < 0)
        return -1;

    if (writes(con, dev_p) < 0)
        return -1;
    if (reads(con, buf, 10) < 0)
        return -1;
    if (*buf != 'O')
    {
        errno = ENOENT;
        return -1;
    }

    if(signal(SIGPIPE, sig_pipe) == SIG_ERR)
    {
        printf("Error installing signal handler\n");
        exit(1);
    }

    mf->sock = con;
    mf->dtype = dtype;
    mf->flags=get_device_flags(name);
    return 0;
}

#define PUTC(c) do {     \
    *p++ = (c);        \
     if (++cnt >= len) { \
         closedir(dir);  \
         return -1;      \
     } } while(0)
#define PUTS(s) do { for(i=0; (s)[i]; i++) PUTC((s)[i]); } while (0)

char* find_tavor(const char* name)
{
  char* ret;
  ret=strstr(name, "23108"); if (ret) return ret;
  ret=strstr(name, "25208"); if (ret) return ret;
  ret=strstr(name, "25218"); if (ret) return ret;
  ret=strstr(name, "25228"); if (ret) return ret;
  ret=strstr(name, "25238"); if (ret) return ret;
  ret=strstr(name, "24204"); if (ret) return ret;
  ret=strstr(name, "25204"); if (ret) return ret;

  // ConnectX
  ret=strstr(name, "25408"); if (ret) return ret;
  ret=strstr(name, "25418"); if (ret) return ret;
  ret=strstr(name, "25438"); if (ret) return ret;
  ret=strstr(name, "25448"); if (ret) return ret;
  ret=strstr(name, "25458"); if (ret) return ret;
  ret=strstr(name, "25468"); if (ret) return ret;
  ret=strstr(name, "25488"); if (ret) return ret;
  ret=strstr(name, "25498"); if (ret) return ret;
  ret=strstr(name, "26418"); if (ret) return ret;
  ret=strstr(name, "26428"); if (ret) return ret;
  ret=strstr(name, "26448"); if (ret) return ret;
  ret=strstr(name, "26478"); if (ret) return ret;

  // Infiniscale4
  ret=strstr(name, "48436"); if (ret) return ret;
  ret=strstr(name, "48437"); if (ret) return ret;
  ret=strstr(name, "48438"); if (ret) return ret;

  return ret;
}


static int get_device_flags(const char *name)
{
  int mask=0;
  char* p;
  if ( find_tavor(name) && strstr(name, "pci_ddr") ) mask|=MDEVS_TAVOR_DDR;
  if ( find_tavor(name) && strstr(name, "pci_uar") ) mask|=MDEVS_TAVOR_UAR;
  if ( find_tavor(name) && strstr(name, "pci_cr") ) mask|=MDEVS_TAVOR_CR;
  if ( strstr(name, "_pciconf") ) mask|=MDEVS_TAVOR_CR;
  if ( strstr(name, "mt21108") ) mask|=MDEVS_GAMLA;
  if ( strstr(name, "vtop")  ) mask|=MDEVS_MEM;
  if ( strstr(name, "calbr") ) mask|=MDEVS_I2CM;
  if ( strstr(name, "calibre") ) mask|=MDEVS_I2CM;
  if ( strstr(name, "mtusb") ) mask|=MDEVS_I2CM;
  if ( strstr(name, "mif") ) mask|=MDEVS_IF;
  if ( strstr(name, "dev-i2c") ) mask|=MDEVS_DEV_I2C;
  if ( strstr(name, "ppc") ) mask|=MDEVS_PPC;
  if ( strstr(name, "i2cm") ) mask|=MDEVS_I2CM;
  if ( strstr(name, ":") ) mask|=MDEVS_REM;
  p = strstr(name, "lid-");
  if ( p ) {
      char *ep;
      strtoul(p + 4, &ep, 0);
      if (p[4] != '\0' && (*ep == '\0' || *ep == ',')) {
          mask |= MDEVS_IB;
      }
  }
  return mask;
}

//
// Hermon PCI ordering bug work around (in user level):
//
// Toggle array:
// Each entry in the array holds the owning PID in bits 30:0 and toggle bit in bit 31.
// The size of the array is the size of max opened mfiles that can be opened simultaniously
// of the machine.
//
// Open:
//   If the wa is enabled ("MTCR_HERMON_WA" env variable does NOT exist or exist and equals 1):
//   hermon_wa_find_slot:
//     for each entry in the wa array:
//       if entry==0 , return it.
//     if all entries != 0, hermon_wa_clear_slots() - check the validity of the PIDs and clear accordingly.
//     if all entries are still != 0, return fail.
//   put the slot addr in the mfile struct.
//   zero the toggle bit in crspace.
//   zero the toggle bit in the mfile.
//
//   NOTE: The open in this implementation is not safe - if 2 processes run hermon_wa_find_slot() in the same time,
//         they can both get the same slot.
//         Maybe use a HW semaphore to protect hermon_wa_find_slot() ?
//
// Write:
//   write to  the given register
//   flip the toggle bit im the mfile struct
//   write the toggle bit the the wa slot addr in crspace.
//   poll read the wa slot untill the toggled result is read.
//   rewrite the new val after every poll read - This was added because polling by reads only sometimes
//     hangs forever. The writes seem to flush the fifos in TGU - need to investigate this.
//
// Close:
//   Zero the slot.
//


//
// Hermon PCI ordering bug work around (in kernel):
//
// Toggle bit is done using a HW semaphore. 3 slots (sems) are allocated for MST tools and
// are dynamically allocated by the mst_pci.o module. 1 slot is reserved for external tools usage.
// This means to more than 3 tools that use the PCI dev can run simultaniously.
//
// Open:
//   If the wa is enabled ("MTCR_HERMON_WA" env variable does NOT exist or exist and equals 1):
//   and HERMON-A0 dev id is read:
//   get a vacant slot - ioctl( , PCI_HERMON_WA ,);
//   read the sem in the slot (to lock it).
//
//
// Flush (hermon_wa_write_confirm_sem):
//   If enabled on first read after write:
//   write 0 to sem.
//   read sem untlill sem = 0;
//
// Close/release
//   Zero the slot.
//

void hermon_wa_clear_stots(mfile* mf) {
    // kill(0) all the !-0 pids in the array.
    // If the pid does not exist - clear the entry.
    int i;

    for (i = 0; i < HERMON_WA_SIZE; i++ ) {
        u_int32_t slot_value;
        u_int32_t slot_addr =  HERMON_WA_BASE + i*4;
        u_int32_t pid;

        if (mread4(mf, slot_addr, &slot_value) != 4) return;
        pid = slot_value & 0x7fffffff;

        if (pid) {
            // check if this pid exists:
            if ( kill(pid, 0) ) {
                // kill failed - no such process.- zero this entry.
                *((u_int32_t *)((char *)mf->ptr + slot_addr)) = 0;
            }
        }
    }
}

// Returns cr address of the slot to use, of 0 if no vacant slot is found.
u_int32_t hermon_wa_find_slot(mfile* mf) {
    int i;

    for (i = 0; i < HERMON_WA_SIZE; i++) {
        u_int32_t slot_value;
        u_int32_t slot_addr = HERMON_WA_BASE + i*4;

        if (mread4(mf, slot_addr, &slot_value) != 4) return 0;
        if (slot_value == 0) {
            // printf("-D- Hermon WA: Found slot: %06x\n", slot_addr);
            return slot_addr;
        }
    }
    printf("-DE- Hermon WA: No empty slot found. Base = %06x, %d slots\n", HERMON_WA_BASE, HERMON_WA_SIZE);
    return 0;
}

// TODO: addr is given for debug. Can remove for perf later.

int hermon_wa_write_confirm_sem(mfile* mf, u_int32_t addr) {
    u_int64_t max_retries = 0x1000000;
    u_int64_t retries = 0;

    mf->hermon_wa_num_of_writes++;

    // clear sem
    *((u_int32_t *)((char *)mf->ptr + mf->hermon_wa_slot)) = 0;

    while (retries < max_retries) {
         u_int32_t slot_value ;

         slot_value = __be32_to_cpu(*((volatile u_int32_t *)((char *)mf->ptr + mf->hermon_wa_slot)));
         if (slot_value == 0) {

             // Good - The written value was read
             if (retries) {
                 // NOT REALLY NEEDED - just to collect statistics

                 // printf("-D- Hermon WA addr %06x took %d retries\n", addr, retries);
                 mf->hermon_wa_num_of_retry_writes++;
                 if (retries > mf->hermon_wa_max_retries) {
                     mf->hermon_wa_max_retries = retries;
                 }
             }
             return 4;
         } else {
             // sem is still locked from previous read - read passed the write
             retries++;
         }
     }

     printf("-D- Hermon WA addr %06x failed after %ld retries\n", addr, retries);
     return 0;
}

int hermon_wa_write_confirm(mfile* mf, u_int32_t addr) {
    u_int64_t max_retries = 0x1000000;
    u_int64_t retries = 0;

    mf->hermon_wa_num_of_writes++;

    mf->hermon_wa_stat ^= 0x80000000;
    *((u_int32_t *)((char *)mf->ptr + mf->hermon_wa_slot)) = __cpu_to_be32(mf->hermon_wa_stat);

    while (retries < max_retries) {
        u_int32_t slot_value ;

        slot_value = __be32_to_cpu(*((volatile u_int32_t *)((char *)mf->ptr + mf->hermon_wa_slot)));
        //mread4(mf, mf->hermon_wa_slot, &slot_value);
        //printf("-D- Hermon WA addr %06x: retry %9d: stat: %08x slot: %08x\n", addr, retries, mf->hermon_wa_stat, slot_value);
        if (slot_value == mf->hermon_wa_stat) {

            // Good - The written value was read
            if (retries) {
                // NOT REALLY NEEDED - just to collect statistics
                // printf("-D- Hermon WA addr %06x took %d retries\n", addr, retries);
                mf->hermon_wa_num_of_retry_writes++;
                if (retries > mf->hermon_wa_max_retries) {
                    mf->hermon_wa_max_retries = retries;
                }
            }
            return 4;
        }

        retries++;
    }

    printf("-D- Hermon WA addr %06x failed after %ld retries\n", addr, retries);
    return 0;
}

#define SEM_ERROR (-2)

#ifdef USE_I2C_SEM


#include <semaphore.h>
#include <time.h>
#include <libgen.h>

// Open the sem and place it in the generic ptr field.
static int i2c_sem_open(mfile* mf, const char* name) {
    char* semname = SEMNAME(name);
    sem_t* psem;
    char* i2c_sem_env = getenv("MTCR_I2C_SEM");
    if (i2c_sem_env != NULL && !strcmp(i2c_sem_env, "0")) {
        return 0;
    }

    sem_t* psem = sem_open(semname, 0);
    if (SEM_FAILED == psem) {
        printf("-D- sem_open: %s:\n", strerror(errno));
        return -1;
    }
    //printf("-D- %8d: sem_opened %s\n", getpid(), semname);
    mf->ptr = psem;
    return 0;
}

static int i2c_sem_close(mfile* mf) {
    if (mf->ptr) {
        sem_close(mf->ptr);
        //printf("-D- %8d: sem_oclosed\n", getpid());

    }
    return 0;
}

static int i2c_sem_wait(mfile* mf) {
    sem_t* sem = (sem_t*)(mf->ptr);
    if (sem) {
        int rc;
        struct timespec ts;
        if (clock_gettime(CLOCK_REALTIME, &ts) == -1)
            return SEM_ERROR;
        ts.tv_sec += 4; // 4 secs is good enough

        rc = sem_timedwait(sem, &ts);
        if (rc) {
            return SEM_ERROR;
        }
        //printf("-D- %8d: sem_locked\n", getpid());
        //getchar();
    }

    return 0;
}

static int i2c_sem_post(mfile* mf) {
    sem_t* sem = (sem_t*)(mf->ptr);
    if (sem) {
        sem_post(sem);
        //printf("-D- %8d: sem_released\n", getpid());
    }
    return 0;
}

#elif defined(USE_I2C_LOCK)

#include <sys/file.h>
static int i2c_sem_open(mfile* mf, const char* name) {
    char* i2c_sem_env = getenv("MTCR_I2C_SEM");
    name = NULL;

    if (i2c_sem_env != NULL && !strcmp(i2c_sem_env, "0")) {
        mf->ptr = NULL;
    } else {
        mf->ptr = (void*)1;
    }

    return 0;
}

static int i2c_sem_close(mfile* mf) {
    mf = NULL;
    return 0;
}

#define I2C_LOCK_RETRIES (500000)
static int i2c_sem_wait(mfile* mf) {
    int rc;
    int i;

    if (!mf->ptr) {
        return 0;
    }

    for (i = 0; i < I2C_LOCK_RETRIES; i++) {
        rc = flock(mf->fd, LOCK_EX|LOCK_NB);
        if (rc == 0 ) {
            return 0;
        } else if (errno != EWOULDBLOCK ) {
            return -1;
        }
    }
    return SEM_ERROR;
}

static int i2c_sem_post(mfile* mf) {
    int rc;

    if (!mf->ptr) {
        return 0;
    }

    rc = flock(mf->fd, LOCK_UN);
    if (rc != 0) {
        return -1;
    }
    return 0;
}

#else

// Stub sem funcs
static int i2c_sem_open(mfile* mf, const char* name) {
    mf = NULL; name = NULL;
    return 0;
}

static int i2c_sem_close(mfile* mf) {
    mf = NULL;
    return 0;
}

static int i2c_sem_wait(mfile* mf) {
    mf = NULL;
    return 0;
}

static int i2c_sem_post(mfile* mf) {
    mf = NULL;
    return 0;
}

#endif


// return -1 - dimax failed. -2 sem failed.
static int dimax_ReadI2c_sem ( mfile* mf, int fd, PI2C_TRANS tr, int size ) {
    int rc;

    rc = i2c_sem_wait(mf);           if (rc) return rc;
    rc = dimax_ReadI2c(fd, tr, size); if (rc) return rc;
    rc = i2c_sem_post(mf);           if (rc) return rc;

    return 0;
}

static int dimax_WriteI2c_sem(mfile* mf, int fd, PI2C_TRANS tr, int size) {
    int rc;

    rc = i2c_sem_wait(mf);           if (rc) return rc;
    rc = dimax_WriteI2c(fd, tr, size);if (rc) return rc;
    rc = i2c_sem_post(mf);           if (rc) return rc;

    return 0;
}

int mdevices(char *buf, int len, int mask)
{
    DIR           *dir;
    struct dirent *dirent;
    int           i, rc=0, cnt=0;
    char         *p = buf;

    /* Regular devices */
    if ((dir = opendir(MST_DIR)) != NULL)
    {

        while ((dirent = readdir(dir)) != NULL)
        {
            char *name = dirent->d_name;

            /* Skip "." and ".." */
            if (name[0] == '.')
                continue;

            /* According to mask */
            if (get_device_flags(name)&mask)
            {
                PUTS(MST_DIR);
                PUTC('/');
                PUTS(name);
                PUTC('\0');
                rc++;
            }
        }
        closedir(dir);
    }

    return rc;
}

mfile *mopen(const char *name)
{
    return mopend(name, MST_TAVOR);
}

mfile *mopend(const char *name, DType dtype)
{
    int  err;
    char *real_name = (char *)name;
    char *p;


    /*  Local open */
    /* allocate mfile struct */
    mfile *mf = malloc(sizeof(mfile));

    if (!mf)
    {
        errno = ENOMEM;
        return 0;
    }


    /*  Remote connection? - Special open */
    if (strchr(name, ':'))
    {
        mf->is_i2cm = 0;
        if (remote_open(name, mf, dtype) < 0)
            return 0;
        else
            return mf;

    }

    mf->sock = -1;

    /*  IB device? special open */
    if ((p = strstr(name, "lid-")) != 0)
    {
        mf->tp = MST_IB;
        if (mib_open(p+4, mf) < 0)
            return 0;
        else
            return mf;
    }

    /* Type of device */
    mf->dtype = dtype;
    if (dtype == MST_TAVOR)
        mf->i2c_slave = 0x48;
    else
        mf->i2c_slave = 0x5a;


    /* Use the device as I2C master? */
    mf->is_i2cm = strstr(name, "i2cm") ? 1 : 0;

    /* Type of interface (relevant when is_i2cm==1 or target==MST_DIMM,
       unused otherwise */
    if (find_tavor(name))
        mf->itype = MST_TAVOR;
    else if (strstr(name, "mt47396"))
        mf->itype = MST_TAVOR;         /*  Anafa-2 like Tavor!!! */
    else if (strstr(name, "mt21108"))
        mf->itype = MST_GAMLA;


    /* Type of file */
    if (strstr(name, "pciconf"))
        mf->tp = MST_PCICONF;
    else if (strstr(name, "pci"))
        mf->tp = MST_PCI;
    else if (strstr(name, "mtusb"))
        mf->tp = MST_USB;
    else if (strstr(name, "calbr") || strstr(name, "calibre"))
        mf->tp = MST_CALBR;
    else if (strstr(name, "mif") != 0)
        mf->tp = MST_IF;
#ifdef ENABLE_MST_DEV_I2C
    else if (strstr(name, "dev-i2c") != 0)
        mf->tp = MST_DEV_I2C;
#endif
    else if (strstr(name, "ppc") != 0)
        mf->tp = MST_PPC;       /*  PPC CPU Interface ( PowerQuick ) */
    else
    {
        free(mf);
        errno = ENOENT;
        return 0;
    }

    /* Open it */
    if ((mf->fd = open(real_name, O_RDWR | O_SYNC)) < 0)
    {
        err = errno;
        free(mf);
        errno = err;
        return 0;
    }


    mf->flags=get_device_flags(name);
    /* Post open actions */
    switch (mf->tp)
    {
    case MST_PCI:
    {
        /* Map memory for PCI interface */
        struct mst_pci_params_st pars;
        int rc;
        u_int32_t devid = 0;
        CLEAR(pars);

        rc=ioctl(mf->fd, PCI_PARAMS, &pars);
        if (rc < 0)
        {
            err = errno;
            close(mf->fd);
            free(mf);
            errno = err;
            return 0;
        }

        mf->ptr = mmap(NULL, pars.size, PROT_READ | PROT_WRITE,
                       MAP_SHARED, mf->fd, 0);

        if (mf->ptr == MAP_FAILED)
        {
            err = errno;
            close(mf->fd);
            free(mf);
            errno = err;
            return 0;
        }

        mf->map_size = (unsigned long)pars.size;

        // Hermon WA setup
        // printf("-D- Hermon WA setup\n");
        mf->hermon_wa_slot = 0;

        if (mf->flags & MDEVS_TAVOR_CR) {

            // Check if use WA
            char* hermon_wa_env = getenv("MTCR_HERMON_WA");
            if (hermon_wa_env == NULL || strcmp(hermon_wa_env, "0")) {
                mread4(mf, 0xf0014, &devid);
                if (devid == 0xa00190) { // TODO: Should I ignore REV here ?

                    // Hermon A0
                    #ifdef HERMON_WA_USERLEVEL
                    // User level WA is not safe if 2 processes open simultaniously
                    mf->hermon_wa_slot = hermon_wa_find_slot(mf);
                    if (mf->hermon_wa_slot == 0) {
                        hermon_wa_clear_stots(mf);
                        mf->hermon_wa_slot = hermon_wa_find_slot(mf);

                        if (mf->hermon_wa_slot == 0) {
                            err = errno;
                            close(mf->fd);
                            free(mf);
                            errno = err;
                            return 0;
                        }
                    }

                    mf->hermon_wa_stat = getpid();

                    #else

                    unsigned int slot_num;

                    rc=ioctl(mf->fd, PCI_HERMON_WA, &slot_num);
                    if (rc < 0)
                    {
                        err = errno;
                        close(mf->fd);
                        free(mf);
                        errno = err;
                        return 0;
                    }

                    mf->hermon_wa_slot = HERMON_WA_BASE + 4*slot_num;

                    mf->hermon_wa_max_retries = 0;
                    mf->hermon_wa_num_of_writes = 0;
                    mf->hermon_wa_num_of_retry_writes = 0;

                    hermon_wa_write_confirm_sem(mf, mf->hermon_wa_slot);
                    #endif
                }
            }
        }

        break;
    }
    case MST_CALBR:
    /* { */
     /* Dummy transaction (first transaction may fail) */
    /*     u_int32_t val; */
    /*     mread4(mf, mf->dtype == MST_TAVOR ? 0xf0014 : 0x2800, &val); */
          break;
    /* } */
    case MST_PCICONF:
          break;

#ifndef MST_CPU_ppc_8xx
    case MST_USB:
    {
      int rc;
      struct usb_device_descriptor desc;
      rc=usb_get_device_descriptor(mf->fd, &desc);
      if (rc) {
        err = errno;
        close(mf->fd);
        free(mf);
        errno = err;
        return 0;
      }
      if (desc.idVendor == DEVASYS_V && desc.idProduct == DEVASYS_P)
      {
        mf->tp = MST_USB;
      }
      else if (desc.idVendor ==  DIMAX_V && desc.idProduct == DIMAX_P)
      {
        mf->tp = MST_USB_DIMAX;
      }
      else
      {
        close(mf->fd);
        free(mf);
        errno = ENOTTY;
        return 0;
      }

      if (0 != i2c_sem_open(mf, name)) {
          int en = errno;
          close(mf->fd);
          free(mf);
          errno = en;
          return 0;
      }
    }
#endif
    case MST_IB:
    case MST_IF:
#ifdef ENABLE_MST_DEV_I2C
    case MST_DEV_I2C:
#endif
        break;

    case MST_PPC:
    {
#ifdef MST_CPU_ppc_8xx
        /* Map memory for PCI interface */
        struct mst_ppc_params_st pars;
        CLEAR(pars);

        if (ioctl(mf->fd, PPC_PARAMS, &pars) < 0)
        {
            err = errno;
            close(mf->fd);
            free(mf);
            errno = err;
            return 0;
        }
        mf->ptr = mmap(NULL, pars.size, PROT_READ | PROT_WRITE,
                       MAP_SHARED, mf->fd, 0);
        if (mf->ptr == MAP_FAILED)
        {
            err = errno;
            close(mf->fd);
            free(mf);
            errno = err;
            return 0;
        }
#endif
        break;
    }

    default:
      close(mf->fd);
      free(mf);
      errno = EINVAL;
      return 0;
    }

    /* OK */
    return mf;
}

int mclose(mfile *mf)
{
    int rc;

    if (mf->tp == MST_IB)
        return mib_close(mf);

    if (mf->sock != -1)
    {
        /*  Remote connection */
        writes(mf->sock, "C");
        rc = close(mf->sock);
        mf->sock = -1;
    } else {
        if (mf->tp == MST_PCI) {

            if (mf->hermon_wa_slot) {
                // *((u_int32_t *)((char *)mf->ptr + mf->hermon_wa_slot)) = 0;
                if (getenv("MTCR_HERMON_WA")) {
                    printf("-D- Hermon WA stats:\n");
                    printf("-D- : num of write flushes: %8ld\n", mf->hermon_wa_num_of_writes);
                    printf("-D- : num of retry flushes: %8ld\n", mf->hermon_wa_num_of_retry_writes);
                    printf("-D- : max_retries:          %8ld\n", mf->hermon_wa_max_retries);
                }
            }

            munmap(mf->ptr, mf->map_size);
        } else if (mf->tp == MST_USB) {
            i2c_sem_close(mf);
        }

        rc = close(mf->fd);
    }

    free(mf);
    return rc;
}

void maccelerate(mfile *mf)
{
    if (mf->is_i2cm)
        i2c_master_set(mf);
}

void mrestore(mfile *mf)
{
    if (mf->is_i2cm)
        i2c_master_restore(mf);
}

static
int prepare_i2c_buf(void* maddr, DType dtype, u_int32_t offset)
{
        switch(dtype)
        {
        case MST_TAVOR:
        {
            u_int32_t offs = __cpu_to_be32(offset);
            memcpy(maddr, &offs, 4);
            return 4;
        }
        case MST_GAMLA:
        {
            u_int16_t offs = offset & 0xffff;
            offs = __cpu_to_be16(offs);
            memcpy(maddr, &offs, 2);
            return 2;
        }
        case MST_DIMM:
        {
            u_int8_t offs1 = offset & 0xff;
            memcpy(maddr, &offs1, 1);
            return 1;
        }
        default:
            return 0;
        }
}

static inline
int prepare_i2c_data(unsigned char* buf, DType dtype, u_int32_t offset, void* data,
    int data_length)
{
  int len;
  len=prepare_i2c_buf(buf,dtype,offset);
  memcpy(buf+len,data, data_length);
  return data_length+len;
}

int mread4(mfile *mf, unsigned int offset, u_int32_t *value)
{
    int rc;
    mif_param_t param;

    if (offset & 0x3) {
        errno = EINVAL;
        return -1;
    }

    if (mf->sock != -1)
    {
        /*  Remote connection */
        char buf[20];

        sprintf(buf, "R 0x%08x", offset);
        writes(mf->sock, buf);
        reads(mf->sock, buf, 20);
        if (*buf == 'O')
        {
            *value = strtoul(buf+2, 0, 0);
            return 4;
        }
        else
        {
            errno = EIO;
            return -1;
        }
    }

    switch(mf->tp)
    {
    case MST_IB:
        return mib_read4(mf, offset, value);

    case MST_IF:
        param.cmd = MREAD4;
        param.buf = value;
        param.len = 4;
        param.addr  = mf->i2c_slave;
        param.offset = offset;
        param.dtype = mf->dtype;
        return(ioctl(mf->fd, param.cmd, &param));


#ifdef ENABLE_MST_DEV_I2C
    case MST_DEV_I2C:
    {
        struct i2c_rdwr_ioctl_data i2c_rdwr;
        struct i2c_msg i2c_msg[2];
        char    maddr[4];
        char    data[4];

        i2c_msg[0].addr = mf->i2c_slave;
        i2c_msg[0].flags = 0;
        i2c_msg[0].buf = &maddr[0];
        i2c_msg[1].addr = mf->i2c_slave;
        i2c_msg[1].flags = I2C_M_RD;
        i2c_msg[1].len = 4;
        i2c_msg[1].buf = data;
        i2c_rdwr.msgs = i2c_msg;
        i2c_rdwr.nmsgs = 2;

        i2c_msg[0].len = prepare_i2c_buf(maddr, mf->dtype, offset);

        if (!i2c_msg[0].len)
        {
            i2c_msg[0].flags = I2C_M_RD;
            i2c_msg[0].len = 4;
            i2c_msg[0].buf = data;
            i2c_rdwr.nmsgs = 1;
        }

        rc=ioctl( mf->fd, I2C_RDWR, &i2c_rdwr );
        if( rc<0 ) return rc;
        *value = __be32_to_cpu(*((u_int32_t *)&data[0]));
        return 4;
        return -1;
    }/*MST_DEV_I2C*/
#endif

    case MST_PCI:
    case MST_PPC:
        if (!mf->ptr)
        {
            errno = EFAULT;
            return -1;
        }
        if (mf->is_i2cm)
            return i2c_master_read_cr(mf, value, offset, 4);

        if (mf->itype == MST_TAVOR) {
            #ifndef HERMON_WA_ALL_WRITES
            if (mf->hermon_wa_slot && mf->hermon_wa_last_op_write) {
                if (hermon_wa_write_confirm_sem(mf, offset) == 0) return -1;
                mf->hermon_wa_last_op_write = 0;
            }
            #endif

            *value = __be32_to_cpu(*((u_int32_t *)((char *)mf->ptr + offset)));
        } else
            *value = __le32_to_cpu(*((u_int32_t *)((char *)mf->ptr + offset)));
        break;
    case MST_PCICONF:
    {
        struct mst_read4_st r4;

        if (mf->is_i2cm)
            return i2c_master_read_cr(mf, value, offset, 4);

        CLEAR(r4);
        r4.offset = offset;
        if ((rc = ioctl(mf->fd, PCICONF_READ4, &r4)) < 0)
            return rc;
        *value = r4.data;
        break;
    }
    case MST_USB:
    case MST_USB_DIMAX:
    {
        int i;
        int addr_transaction = 1;

        for (i=0; i<MAX_USB_RD_RETRY; i++)
        {
            I2C_TRANS tr;

            tr.byTransType = I2C_TRANS_NOADR;
            tr.bySlvDevAddr = mf->i2c_slave << 1;
            switch(mf->dtype)
            {
            case MST_TAVOR:
            {
                u_int32_t offs = __cpu_to_be32(offset);
                memcpy(tr.Data, &offs, 4);
                tr.wCount = 4;
                break;
            }
            case MST_GAMLA:
            {
                u_int16_t offs = offset & 0xffff;
                offs = __cpu_to_be16(offs);
                memcpy(tr.Data, &offs, 2);
                tr.wCount = 2;
                break;
            }
            case MST_DIMM:
            {
                u_int8_t offs1 = offset & 0xff;
                memcpy(tr.Data, &offs1, 1);
                tr.wCount = 1;
                break;
            }
            case MST_NOADDR:
                addr_transaction = 0;
                tr.wCount = 4;
                break;

            }

            if( mf->tp == MST_USB_DIMAX )
            {
                /*  Dimax Device */
                if( ! addr_transaction )
                    tr.wCount = 0;
                rc = dimax_ReadI2c_sem(mf, mf->fd, &tr, 4 );
                if( rc ) {
                    if (rc != SEM_ERROR)
                        errno = EIO;
                    return -1;
                }
                *value = __be32_to_cpu(*((u_int32_t *)&tr.Data[0]));
                return 4;
            }
            else
            {
                /*  Write address if neccesary */
                rc = addr_transaction ? mst_WriteI2c(mf->fd, &tr) : 4;
                if (rc > 0)
                {
                    if (rc != tr.wCount)
                    {
                        errno = EIO;
                        return -1;
                    }

                    /*  Read result */
                    tr.wCount = 4;
                    if ((rc = mst_ReadI2c(mf->fd, &tr)) > 0)
                    {
                        if (rc != tr.wCount)
                        {
                            errno = EIO;
                            return -1;
                        }

                        *value = __be32_to_cpu(*((u_int32_t *)&tr.Data[0]));
                        return 4;
                    }
                }
            }
        }
        errno = EPIPE;
        return -1;
    }
    case MST_CALBR:
    {
        struct mst_trans_st tr;
        CLEAR(tr);

        tr.slv_addr = mf->i2c_slave;
        tr.wr_size = prepare_i2c_buf(tr.wr_buff, mf->dtype, offset);
        tr.rd_size = sizeof(u_int32_t);

        if ((rc = ioctl(mf->fd, CALBR_TRANS, &tr)) < 0)
            return -1;
        if (tr.rd_count <  sizeof(u_int32_t))
            return tr.rd_count;
        *value = __be32_to_cpu(*((u_int32_t *)&tr.rd_buff[0]));
        break;
    }
    }
    return 4;
}

int mwrite4(mfile *mf, unsigned int offset, u_int32_t value)
{
    int rc;
    mif_param_t param;

    if (offset & 0x3) {
        errno = EINVAL;
        return -1;
    }

    if (mf->sock != -1)
    {
        /*  Remote connection */
        char buf[40];

        sprintf(buf, "W 0x%08x 0x%08x", offset, value);
        /* printf("W 0x%08x 0x%08x", offset, value); */
        writes(mf->sock, buf);
        reads(mf->sock, buf, 40);
        if (*buf == 'O')
            return 4;
        else
        {
            errno = EIO;
            return -1;
        }
    }

    switch(mf->tp)
    {
    case MST_IB:
        return mib_write4(mf, offset, value);

    case MST_IF:
        param.cmd = MWRITE4;
        param.buf = &value;
        param.len = 4;
        param.addr = mf->i2c_slave;
        param.offset = offset;
        param.dtype = mf->dtype;
        return(ioctl(mf->fd, param.cmd, &param));

#ifdef ENABLE_MST_DEV_I2C
    case MST_DEV_I2C:
    {
        struct i2c_rdwr_ioctl_data i2c_rdwr;
        struct i2c_msg i2c_msg[1];
        unsigned char data[8];

        i2c_msg[0].addr = mf->i2c_slave;
        i2c_msg[0].flags = 0;
        i2c_msg[0].buf = data;
        i2c_rdwr.msgs = i2c_msg;
        i2c_rdwr.nmsgs = 1;

        value = __cpu_to_be32(value);
        i2c_msg[0].len = prepare_i2c_data(data, mf->dtype, offset,
            &value, sizeof(value));

        rc=ioctl( mf->fd, I2C_RDWR, &i2c_rdwr );
        if( rc<0 ) return rc;
        return 4;
    } /*MST_DEV_I2C*/
#endif

    case MST_PCI:
    case MST_PPC:
        if (!mf->ptr)
        {
            errno = EFAULT;
            return -1;
        }
        if (mf->is_i2cm)
            return i2c_master_write_cr(mf, value, offset, 4);

        if (mf->itype == MST_TAVOR) {
            *((u_int32_t *)((char *)mf->ptr + offset)) = __cpu_to_be32(value);
            #ifdef HERMON_WA_ALL_WRITES
            if (mf->hermon_wa_slot) {
                if (hermon_wa_write_confirm_sem(mf, offset) == 0) return -1;
            }
            #else
            mf->hermon_wa_last_op_write = 1;
            #endif
        } else
            *((u_int32_t *)((char *)mf->ptr + offset)) = __cpu_to_le32(value);
        break;
    case MST_PCICONF:
    {
        struct mst_write4_st r4;

        if (mf->is_i2cm)
            return i2c_master_write_cr(mf, value, offset, 4);

        CLEAR(r4);
        r4.offset = offset;
        r4.data = value;
        if ((rc = ioctl(mf->fd, PCICONF_WRITE4, &r4)) < 0)
            return rc;
        break;
    }
    case MST_USB:
    case MST_USB_DIMAX:
    {
        int i;
        for (i=0; i<MAX_USB_RD_RETRY; i++)
        {
            I2C_TRANS tr;

            tr.byTransType = I2C_TRANS_NOADR;
            tr.bySlvDevAddr = mf->i2c_slave << 1;
            value = __cpu_to_be32(value);
            tr.wCount = prepare_i2c_data(tr.Data, mf->dtype, offset,
                &value, sizeof(value));

            if( mf->tp == MST_USB_DIMAX )
            {
                tr.wCount -= sizeof(value);     /*  leave only address size */
                rc = dimax_WriteI2c_sem(mf, mf->fd, &tr, 4 );
                if( rc ) {
                    errno = EIO;
                    return -1;
                }
                return 4;
            }
            else
            {
                /*  Write address and data */
                if ((rc = mst_WriteI2c(mf->fd, &tr)) > 0)
                {
                    if (rc != tr.wCount)
                    {
                        errno = EIO;
                        return -1;
                    }
                    return 4;
                }
            }
        }
        errno = EPIPE;
        return -1;
    }
    case MST_CALBR:
    {
        struct mst_trans_st tr;
        CLEAR(tr);

        tr.slv_addr = mf->i2c_slave;
        value = __cpu_to_be32(value);
        tr.wr_size=prepare_i2c_data( tr.wr_buff, mf->dtype, offset, &value,
            sizeof(value));

        tr.rd_size = 0;

        if ((rc = ioctl(mf->fd, CALBR_TRANS, &tr)) < 0)
            return -1;
        if (tr.wr_count <  sizeof(u_int32_t))
            return tr.wr_count;
        break;
    }
    }
    return 4;
}

int mread64(mfile *mf, unsigned int offset, void *data, int length)
{
    int                 rc;
    mif_param_t         param;

    if (length > MAX_TRANS_SIZE)
    {
        errno = EINVAL;
        return -1;
    }

    switch(mf->tp)
    {
    case MST_IB:
        return mib_read64(mf, offset, data, length);


    case MST_IF:
        param.cmd = MREAD64;
        param.buf = data;
        param.len    = length;
        param.addr   = mf->i2c_slave;
        param.offset = offset;
        param.dtype = mf->dtype;
        return(ioctl(mf->fd, param.cmd, &param));

#ifdef ENABLE_MST_DEV_I2C
    case MST_DEV_I2C:
    {
        struct i2c_rdwr_ioctl_data i2c_rdwr;
        struct i2c_msg i2c_msg[2];
        char    maddr[4];

        i2c_msg[0].addr = mf->i2c_slave;
        i2c_msg[0].flags = 0;
        i2c_msg[0].buf = &maddr[0];
        i2c_msg[1].addr = mf->i2c_slave;
        i2c_msg[1].flags = I2C_M_RD;
        i2c_msg[1].len = length;
        i2c_msg[1].buf = data;
        i2c_rdwr.msgs = i2c_msg;
        i2c_rdwr.nmsgs = 2;
        i2c_msg[0].len=prepare_i2c_buf(maddr,mf->dtype,offset);

        if (!i2c_msg[0].len)
        {
            i2c_msg[0].flags = I2C_M_RD;
            i2c_msg[0].len = length;
            i2c_msg[0].buf = data;
            i2c_rdwr.nmsgs = 1;
        }

        rc=ioctl( mf->fd, I2C_RDWR, &i2c_rdwr );
        if( rc<0 ) return rc;
        return length;
        }/*MST_DEV_I2C*/
#endif

    case MST_CALBR:
    {
        struct mst_trans_st tr;
        CLEAR(tr);

        tr.slv_addr = mf->i2c_slave;
        tr.wr_size =prepare_i2c_buf(tr.wr_buff, mf->dtype, offset);
        tr.rd_size = length;

        if ((rc = ioctl(mf->fd, CALBR_TRANS, &tr)) < 0)
            return -1;
        memcpy(data, tr.rd_buff, tr.rd_count);
        return tr.rd_count;
    }
    case MST_USB:
    case MST_USB_DIMAX:
    {
        int       addr_transaction = 1;
        I2C_TRANS tr;
        CLEAR(tr);

        tr.byTransType = I2C_TRANS_NOADR;
        tr.bySlvDevAddr = mf->i2c_slave << 1;
        tr.wCount = prepare_i2c_buf(tr.Data, mf->dtype, offset);
        if (!tr.wCount)
        {
            addr_transaction = 0;
            tr.wCount = 4;
        }

        if( mf->tp == MST_USB_DIMAX )
        {
            /*  Dimax Device */
            if( addr_transaction == 0 )
                tr.wCount = 0;
            rc = dimax_ReadI2c_sem( mf, mf->fd, &tr, length );
            if( rc ) {
                errno = EIO;
                return -1;
            }
            memcpy(data, &tr.Data[0], length);
            return length;
        }
        else
        {
            /*  Write address */
            rc = addr_transaction ? mst_WriteI2c(mf->fd, &tr) : 4;
            if (rc < 0)
                return -1;
            /*  Read result */
            tr.wCount = length;
            if ((rc = mst_ReadI2c(mf->fd, &tr)) < 0)
                return -1;
            memcpy(data, &tr.Data[0], length);
            return length;
        }
    }
    case MST_PCI:
    case MST_PCICONF:
    {
        int i;
        unsigned char *cdata = (unsigned char *)data;

        for (i=0; i<length; i++)
        {
            rc = i2c_master_read_cr(mf, cdata++, offset++, 1);
            if (rc < 0)
                return rc;
            else if (rc < 1)
                return i;
        }
        return length;
    }
    case MST_PPC:
        errno = ECANCELED;
        return -1;
    }

    errno = EPERM;
    return -1;
}

int mwrite64(mfile *mf, unsigned int offset, void *data, int length)
{
    int                 rc;
    mif_param_t param;

    if (length > MAX_TRANS_SIZE)
    {
        errno = EINVAL;
        return -1;
    }


    switch(mf->tp)
    {
    case MST_IB:
        return mib_write64(mf, offset, data, length);


    case MST_IF:
        param.cmd = MWRITE64;
        param.buf = data;
        param.offset= offset;
        param.len   = length;
        param.addr  = mf->i2c_slave;
        param.dtype = mf->dtype;
        return(ioctl(mf->fd, param.cmd, &param));

#ifdef ENABLE_MST_DEV_I2C
    case MST_DEV_I2C:
    {
        struct i2c_rdwr_ioctl_data i2c_rdwr;
        struct i2c_msg i2c_msg[1];
        unsigned char buf[64+4];
        i2c_msg[0].addr = mf->i2c_slave;
        i2c_msg[0].flags = 0;
        i2c_msg[0].buf = buf;
        i2c_rdwr.msgs = i2c_msg;
        i2c_rdwr.nmsgs = 1;

        i2c_msg[0].len = prepare_i2c_data(buf, mf->dtype, offset, data, length);
        rc=ioctl( mf->fd, I2C_RDWR, &i2c_rdwr );
        if( rc<0 ) return rc;
        return length;
    }/*MST_DEV_I2C*/
#endif

    case MST_CALBR:
    {
        struct mst_trans_st tr;
        int                 offs_size=0;
        CLEAR(tr);

        tr.slv_addr = mf->i2c_slave;
        tr.wr_size = prepare_i2c_data( tr.wr_buff, mf->dtype, offset, data,
            length);

        tr.rd_size = 0;

	offs_size=tr.wr_size-length;

        if ((rc = ioctl(mf->fd, CALBR_TRANS, &tr)) < 0)
            return -1;
        return (tr.wr_count < offs_size) ? 0 : tr.wr_count - offs_size;
    }
    case MST_USB:
    case MST_USB_DIMAX:
    {
        I2C_TRANS tr;
        CLEAR(tr);

        tr.byTransType = I2C_TRANS_NOADR;
        tr.bySlvDevAddr = mf->i2c_slave << 1;
        tr.wCount = prepare_i2c_data(tr.Data, mf->dtype, offset, data, length);

        if( mf->tp == MST_USB_DIMAX )
        {
            tr.wCount -= length;        /*  leave only address size */
            rc = dimax_WriteI2c_sem(mf, mf->fd, &tr, length );
            if( rc ) {
                errno = EIO;
                return -1;
            }
            return length;
        }

        /*  Write address and data */
        if ((rc = mst_WriteI2c(mf->fd, &tr)) < 0)
            return -1;
        return length;
    }
    case MST_PCI:
    case MST_PCICONF:
    {
        int i;
        unsigned char *cdata = (unsigned char *)data;

        for (i=0; i<length; i++)
        {
            rc = i2c_master_write_cr(mf, *cdata++, offset++, 1);
            if (rc < 0)
                return rc;
            else if (rc < 1)
                return i;
        }
        return length;
    }
    case MST_PPC:
        errno = ECANCELED;
        return -1;

    }

    errno = EPERM;
    return -1;
}

MTCR_API int msw_reset(mfile *mf) {

    switch(mf->tp)
    {
    case MST_IB:
        return mib_swreset(mf);
    default:
        errno = EPERM;
        return -1;
    }
}


static void buff_cpu2be(void *buff, int byte_len) {
    // TODO: Can return with no loop if on be machine
    int i;
    for (i = 0; i < byte_len ; i += 4) {
        u_int32_t val = *((u_int32_t*)(buff + i));
        val = __cpu_to_be32(val);
        *((u_int32_t*)(buff + i)) = val;
    }
}

static int
mread_chunk_as_multi_mread4(mfile *mf, unsigned int offset, void *data, int length)
{
    int i;
    if (length % 4) {
        return EINVAL;
    }
    for (i = 0; i < length ; i += 4) {
        u_int32_t value;

        if (mread4(mf, offset + i, &value) != 4) {
            return -1;
        }
        memcpy(data + i , &value,4);
    }
    return length;
}

//Read up to length
//Length is guaranteed to be up to what get_chunk_size returns
static int
mread_chunk(mfile *mf, unsigned int offset, void *data, int length)
{
    int                 rc;
    mif_param_t         param;

    if (mf->sock != -1) {
        return mread_chunk_as_multi_mread4(mf,offset,data,length);
    }

    if (mf->is_i2cm) return i2c_master_read(mf, data, offset, length);

    switch(mf->tp)
    {
    case MST_IB:
        //return mib_read64(mf, offset, data, length);
        return mread_chunk_as_multi_mread4(mf,offset,data,length);

    case MST_IF:
        param.cmd = MREAD64;
        param.buf = data;
        param.len    = length;
        param.addr   = mf->i2c_slave;
        param.offset = offset;
        param.dtype = mf->dtype;
        return(ioctl(mf->fd, param.cmd, &param));

#ifdef ENABLE_MST_DEV_I2C
    case MST_DEV_I2C:
        return 0;
#endif

    case MST_CALBR:
        {
            struct mst_trans_st tr;
            CLEAR(tr);

            tr.slv_addr = mf->i2c_slave;
            tr.wr_size = prepare_i2c_buf(tr.wr_buff, mf->dtype, offset);
            tr.rd_size = length;

            if ((rc = ioctl(mf->fd, CALBR_TRANS, &tr)) < 0)
                return -1;
            memcpy(data, tr.rd_buff, tr.rd_count);
            buff_cpu2be(data, tr.rd_count);
            return tr.rd_count;
        }
    case MST_USB:
    case MST_USB_DIMAX:
        {
            int i;
            int       addr_transaction = 1;
            for (i=0; i<MAX_USB_RD_RETRY; i++)
            {
                I2C_TRANS tr;
                CLEAR(tr);

                tr.byTransType = I2C_TRANS_NOADR;
                tr.bySlvDevAddr = mf->i2c_slave << 1;
                tr.wCount = prepare_i2c_buf(tr.Data, mf->dtype, offset);
                if (!tr.wCount)
                {
                    addr_transaction = 0;
                    tr.wCount = 4;
                }

                if( mf->tp == MST_USB_DIMAX )
                {
                    /*  Dimax Device */
                    if( addr_transaction == 0 )
                        tr.wCount = 0;
                    rc = dimax_ReadI2c_sem( mf, mf->fd, &tr, length );
                    if( rc ) {
                        errno = EIO;
                        return -1;
                    }
                    memcpy(data, &tr.Data[0], length);
                    buff_cpu2be(data, length);
                    return length;
                }
                else
                {
                    /*  Write address */
                    rc = addr_transaction ? mst_WriteI2c(mf->fd, &tr) : 4;
                    if (rc < 0)
                        return -1;
                    /*  Read result */
                    tr.wCount = length;
                    if ((rc = mst_ReadI2c(mf->fd, &tr)) < 0)  {
                        errno = EIO;
                        return -1;
                    }
                    memcpy(data, &tr.Data[0], length);
                    buff_cpu2be(data, length);
                    return length;
                }
            }
            errno = EPIPE;
            return -1;
        }
    case MST_PCI:
    case MST_PPC:
    case MST_PCICONF:
        {
            //memcpy(data,(char *)mf->ptr + offset,length);
            //return length;
            return mread_chunk_as_multi_mread4(mf,offset,data,length);
        }

    default:
        errno = EPERM;
        return -1;
    }
}

static int
mwrite_chunk_as_multi_mwrite4(mfile *mf, unsigned int offset, void *data, int length)
{
    int i;
    if (length % 4) {
        return EINVAL;
    }
    for (i = 0; i < length ; i += 4) {
        u_int32_t value;
        memcpy(&value,data + i ,4);
        //value = __cpu_to_be32(value);
        if (mwrite4(mf, offset + i, value) != 4) {
            return -1;
        }
    }
    return length;
}

static int
mwrite_chunk(mfile *mf, unsigned int offset, void *data, int length)
{
    int                 rc;

    // Just implement as a series of mwrites:
    if (mf->sock != -1 ||
        mf->is_i2cm
        ) {
        return mwrite_chunk_as_multi_mwrite4(mf,offset,data,length);
    }

    switch (mf->tp) {
    case MST_IB:
    case MST_IF:
    case MST_PCICONF:
    case MST_PPC:
    case MST_PCI:
        return mwrite_chunk_as_multi_mwrite4(mf,offset,data,length);

#ifdef ENABLE_MST_DEV_I2C
    case MST_DEV_I2C:
        return 0;
#endif

    case MST_CALBR:
        {
            struct mst_trans_st tr;
            CLEAR(tr);

            tr.slv_addr = mf->i2c_slave;
            buff_cpu2be(data, length);
            tr.wr_size=prepare_i2c_data( tr.wr_buff, mf->dtype, offset, data,
                                         length);

            tr.rd_size = 0;

            if ((rc = ioctl(mf->fd, CALBR_TRANS, &tr)) < 0)
                return -1;
            if (tr.wr_count < length)
                return tr.wr_count;

            return tr.wr_count;
        }
    case MST_USB:
    case MST_USB_DIMAX:
        {
            buff_cpu2be(data, length);
            return mwrite64(mf, offset, data, length);
        }
    //case MST_PCI:
    //case MST_PPC:
    //    {
    //        printf("-D- memcpy(%p + %x, %p, %x);\n", mf->ptr, offset,data,length);
    //        memcpy((char *)mf->ptr + offset,data,length);
    //        return length;
    //    }

    default:
        errno = EPERM;
        return -1;
    }
}

static int get_chunk_size(mfile *mf)
{
	/*  Remote connection */
    if (mf->sock != -1) return 4;
    switch(mf->tp)
    {
    //case MST_IB:
    case MST_IF:
    case MST_CALBR:
    case MST_USB:
        return 64;
    case MST_USB_DIMAX:
        return 256; //DIMAX FW uses 8-bit counters, thats why.
#ifdef ENABLE_MST_DEV_I2C
    case MST_DEV_I2C:
        return 0;   /* Will be 8K after it will be implemented in mread_by_chunks */
#endif
    case MST_PCI:
        return 256;
    default:
	//TODO: for i2c masters we can also
	//implement speedups by doing bursts
	return 4;
    }
}

int mread4_block (mfile *mf, unsigned int offset, u_int32_t* data, int byte_len) {
    int chunk_size;
    int left_size;
    int bytes;
    void* dest_ptr=data;

    chunk_size=get_chunk_size(mf);

    for (left_size=byte_len;left_size>0;left_size-=chunk_size)
    {
        int toread;
        toread=(left_size>=chunk_size)?chunk_size:left_size;
        bytes=mread_chunk (mf, offset, dest_ptr, toread);
        if (bytes!=toread) {
            return byte_len-left_size;
        }

        offset += chunk_size;
        dest_ptr+=chunk_size;
    }
    return byte_len;

}

int mwrite4_block(mfile *mf, unsigned int offset, u_int32_t* data, int byte_len) {
    int chunk_size;
    int left_size;
    int bytes;
    void* dest_ptr = data;

    chunk_size=get_chunk_size(mf);

    for (left_size=byte_len;left_size>0;left_size-=chunk_size)
    {
        int towrite;
        towrite=(left_size>=chunk_size)?chunk_size:left_size;
        bytes=mwrite_chunk (mf, offset, dest_ptr, towrite);
        if (bytes!=towrite) {
            return byte_len-left_size;
        }

        offset += chunk_size;
        dest_ptr+=chunk_size;
    }
    return byte_len;
}

/*
 * Get signal and do nothing.
 *
 */
static void
sig_pipe(int signo)
{
    printf("Got SIGPIPE...\n");
    return;
}

unsigned char mset_i2c_slave(mfile *mf, unsigned char new_i2c_slave)
{
    unsigned char ret;
    if (mf)
    {
       ret = mf->i2c_slave;
       mf->i2c_slave = new_i2c_slave;
    }
    else
       ret = 0xff;
    return ret;
}

void *
mget_free_pages(mfile *mf, unsigned int order)
{
    unsigned page_size = 4096;

    if (mf->sock != -1)
    {
        /* unsupported on remote connection */
        errno = -EOPNOTSUPP;
        return 0;
    }

    switch(mf->tp)
    {
    case MST_PPC:
        {
            unsigned int *addr;
            addr = mmap (0, page_size << order, PROT_READ | PROT_WRITE,
                        MAP_SHARED, mf->fd, page_size);

            if (addr == MAP_FAILED)
                    return 0;
            else
                    return addr;
        }

    default:
        errno = -EOPNOTSUPP;
        return 0;
    }
}

int
mfree_pages(mfile *mf, void *addr, unsigned int order)
{
    unsigned page_size = 4096;

    if (mf->sock != -1)
    {
        /* unsupported on remote connection */
        errno = -EOPNOTSUPP;
        return -1;
    }

    switch (mf->tp)
    {
    case MST_PPC:
        {
            return munmap (addr, page_size << order);
        }

    default:
        errno = -EOPNOTSUPP;
        return -1;
    }
}

unsigned long
mvtop (mfile *mf, void *va)
{
#ifdef MST_CPU_ppc_8xx
    struct mst_ppc_vtop vtop;

    vtop.va = va;
#endif

    switch(mf->tp)
    {
#ifdef MST_CPU_ppc_8xx
    case MST_PPC:
        if (ioctl (mf->fd, PPC_VTOP, &vtop) < 0) {
                return 0;
        } else {
                return vtop.pa;
        }
#endif
    default:
        errno = -EOPNOTSUPP;
        return 0;
    }
}


