/*
 *
 *  mtserver.c - Mellanox Software tools (mst) server (remote mtcr calls)
 *
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

#include "mtcr.h"
#include "tcp.h"

/*
 * Constants
 */
#define DEF_PORT 23108
#define BUF_LEN  256
#define DEV_LEN  2048

int sdebug = 0;
int port = DEF_PORT;    /* Default port */

#ifdef SIMULATOR

#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#define FILE_PATH "/tmp/mmap.log"
#define NUM_INTS  (0x40000)
#define FILE_SIZE (NUM_INTS * sizeof(int))

u_int32_t* cr_space;
char sim_str[100] = "\t-i[d]  <id>   - set the device id.\n";
int id;
int fd;
mfile *mopen(const char *name)
{
    name = NULL;
    return (mfile*)0;

}


mfile *mopend(const char *name, DType dtype)
{
    name = NULL;
    dtype = 0;
    return (mfile*)1;
}

int mclose(mfile *mf)
{
    mf = NULL;
    return 0;
}

int mdevices(char *buf, int len, int mask)
{   
    len = 0;
    mask = 0;
    strcpy(buf, "Simulator");
    return 1;
}

int mread4(mfile *mf, unsigned int offset, u_int32_t *value)
{
    mf = NULL;
    *value = cr_space[offset/4];
    return 4;
}

int mwrite4(mfile *mf, unsigned int offset, u_int32_t value)
{
    mf = NULL;
    cr_space[offset/4] = value;
    return 4;
}

int check_id_arg(char* av[], int ac, int *i)
{
    char *end;
    if (++(*i) >= ac)
    {
        printf("After switch \"%s\" device id is expected.\n", av[--(*i)]);
        printf("Type \"%s -h\" for help.\n", av[0]);
        exit(1);
    }
    id = (int)strtol(av[*i], &end, 0);
    if (*end)
    {
        printf("Invalid id: \"%s\" -- ?\n", end);
        printf("Type \"%s -h\" for help.\n", av[0]);
        exit(1);
    }
    return 0;

}

int prepare_the_map_file(void)
{
    int i;
	int result;
    fd = open(FILE_PATH, O_RDWR | O_CREAT | O_TRUNC, (mode_t)0600);

    if (fd == -1) {
        printf("-E- Error opening file for writing");
        exit(1);
    }

    result = lseek(fd, FILE_SIZE-1, SEEK_SET);
    if (result == -1) {
        close(fd);
        perror("Error calling lseek() to 'stretch' the file");
        exit(1);
    }

    result = write(fd, "", 1);
    if (result != 1) {
        close(fd);
        perror("Error writing last byte of the file");
        exit(1);
    }

    cr_space = mmap(0, FILE_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (cr_space == MAP_FAILED) {
        close(fd);
        perror("Error mmapping the file");
        exit(1);
    }
    for (i = 0; i < NUM_INTS; ++i) {
        cr_space[i] = 0;
    }

    // write id 
    if (id != 0) {
        cr_space[0xf0014/4] = id;
    }
    return 0;
}

int unmap_and_close_file(void)
{
    if (munmap(cr_space, FILE_SIZE) == -1) {
        perror("Error un-mmapping the file");
    }
    close(fd);
    return 0;
}
#else

char sim_str[100] = "";
int check_id_arg(char* av[], int ac, int *i)
{
    ac = 0;
    printf("Invalid switch \"%s\".\n", av[*i]);
    printf("Type \"%s -h\" for help.\n", av[0]);
    exit(1);

}

int prepare_the_map_file(void)
{
    return 0;
}

int unmap_and_close_file(void) 
{
    return 0;

}
#endif
/* ////////////////////////////////////////////////////////////////////// */
/* ////////////////////////////////////////////////////////////////////// */
void usage(const char *s)
{
    printf("Usage:\n\t%s [switches]\n\n", s);
    printf("Switches may be:\n");
    printf("\t-p[ort] <port> - Listen to specify port (default is %d).\n", port);
    printf("\t-d[ebug]       - Print all socket traffic (for debuging only).\n");
    printf(sim_str);
    printf("\t-h[elp]        - Print help message.\n");
    exit (1);
}

/* ////////////////////////////////////////////////////////////////////// */
static void writes_deb(int con, char *s)
{
    writes(con, s);
    if (sdebug)
        printf("-> %s\n", s);
}

/* ////////////////////////////////////////////////////////////////////// */
void write_err(int con)
{
    writen(con, "E ", 2);
    writes(con, strerror(errno));
    if (sdebug)
        printf("-> E %s\n", strerror(errno));
}

/* ////////////////////////////////////////////////////////////////////// */
void write_ok(int con)
{
    writes_deb(con, "O");
}

/* ////////////////////////////////////////////////////////////////////// */
void mySignal()
{
    exit(0);
}

/* ////////////////////////////////////////////////////////////////////// */
#define CHK(f)    do { if ((f) < 0) { perror(""); exit(1); } } while(0)
#define CHK2(f,m) do { if ((f) < 0) { perror(m); exit(1); } } while(0)
int main (int ac, char *av[])
{
    char *end;
    char  buf[BUF_LEN], dev_buf[DEV_LEN];
    int   i, con, rc;
    mfile *mf = 0;

    /* Command line parsing. */
    for (i=1; i<ac; i++)
    {   
        switch (*av[i])
        {
        case '-':
            ++av[i];
            if (!strcmp(av[i], "p")  ||  !strcmp(av[i], "port"))
            {
                if (++i >= ac)
                {
                    printf("After switch \"%s\" port number is expected.\n",av[--i]);
                    printf("Type \"%s -h\" for help.\n", av[0]);
                    exit(1);
                }
                port = (int)strtol(av[i], &end, 0);
                if (*end)
                {
                    printf("Invalid port: \"%s\" -- ?\n", end);
                    printf("Type \"%s -h\" for help.\n", av[0]);
                    exit(1);
                }
            }
            else if (!strcmp(av[i], "d")  ||  !strcmp(av[i], "debug"))
            {
                sdebug = 1;
            }
            else if (!strcmp(av[i], "h")  ||  !strcmp(av[i], "help"))
            {
                usage(av[0]);
            }
            else if (!strcmp(av[i], "i")  ||  !strcmp(av[i], "id")) {
                check_id_arg(av, ac, &i);
            }
            else
            {
                printf("Invalid switch \"%s\".\n", av[i]);
                printf("Type \"%s -h\" for help.\n", av[0]);
                exit(1);
            }
            break;
        case '?':
            usage(av[0]);
            break;
        default:
            printf("Invalid parameter \"%s\".\n", av[i]);
            printf("Type \"%s -h\" for help.\n", av[0]);
            exit(1);
        }
    }

    prepare_the_map_file();
    signal(SIGPIPE, mySignal);

    /* Now open and start work */
    logset(1);
    con = open_serv_connection(port);
    CHK2(con, "Open connection (server side)");
    for (;;)
    {
        memset(buf, 0, BUF_LEN);
        rc = reads(con, buf, BUF_LEN);
        CHK(rc);
        if (!rc)
            break;   /*  EOF */
        if (sdebug)
            printf("<- %s\n", buf);
        switch(*buf)
        {
        case 'O':   /*  Open mfile */
            if (mf)
                writes_deb(con, "E Already opened");
            else
            {
                DType dtype = strtoul (buf+2, &end, 0);
                if (*end != ' ')
                    /*  Old style (O DEV_NAME) */
                    mf = mopen(buf+2);
                else
                    /*  New style (O FLAG DEV_NAME) */
                    mf = mopend (end+1, dtype);

                if (mf)
                    write_ok(con);
                else
                    write_err(con);
            }
            break;
        case 'C':  /*  Close mfile */
            if (!mf)
                writes_deb(con, "E Not opened");
            else
            {
                if (mclose(mf) < 0)
                    write_err(con);
                else
                {
                    write_ok(con);
                    mf = 0;
                }
            }
            break;
        case 'L':   /*  Get devices list */
            rc = mdevices(dev_buf, DEV_LEN, MDEVS_ALL);
            if (rc < 0)
                write_err(con);
            else
            {
                char *p = &dev_buf[0], vbuf[16];
                sprintf(vbuf, "O %d", rc);
                writes_deb(con, vbuf);
                for (i = 0; i < rc; i++, p += strlen(p)+1)
                    writes_deb(con, p);
            }
            break;
        case 'R':   /*  Read word */
            if (!mf)
                writes_deb(con, "E Not opened");
            else
            {
                unsigned int offset;
                u_int32_t    value;
                offset = strtoul(buf+2, &end, 0);
                if (*end)
                    writes_deb(con, "E Invalid offset");
                else
                {
                    if (mread4(mf, offset, &value) < 4)
                        write_err(con);
                    else
                    {
                        char vbuf[16];
                        sprintf(vbuf, "O 0x%08x", value);
                        writes_deb(con, vbuf);
                    }
                }
            }
            break;
        case 'W':   /*  Write word */
            if (!mf)
                writes_deb(con, "E Not opened");
            else
            {
                unsigned int offset;
                u_int32_t    value;
                char *p = strchr(buf+2, ' ');
                if (!p)
                    writes_deb(con, "E Invalid format (should be OFFS DATA)");
                else
                {
                    *p = '\0';
                    p++;
                    offset = strtoul(buf+2, &end, 0);
                    if (*end)
                        writes_deb(con, "E Invalid offset");
                    else
                    {
                        value = strtoul(p, &end, 0);
                        if (*end)
                            writes_deb(con, "E Invalid data");
                        else
                        {
                            if (mwrite4(mf, offset, value) < 4)
                                write_err(con);
                            else
                                write_ok(con);
                        }
                    }
                }
            }
            break;
        default:
            writes(con, "E Invalid command");
            if (sdebug)
                printf("-> E Invalid command (len:%d cmd:\"%s\")\n",
                       (int)strlen(buf), buf);
            break;
        }
    }

    unmap_and_close_file();
    return 0;
}
