######################################################################
#  ...
#  NAME
#       ...
#
#  COPYRIGHT
#               - Mellanox Confidential and Proprietary -
#
# Copyright (C) Jan. 2004, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
#
# Except as specifically permitted herein, no portion of the Information,
# including but not limited to object code and source code, may be reproduced,
# modified, distributed, republished or otherwise exploited in any form or by
# any means for any purpose without the prior written permission of Mellanox
# Technologies Ltd. Use of software subject to the terms and conditions
# detailed in the file "LICENSE.txt".
# End of legal section.
#
#  FUNCTION
#     ...
#
#  AUTHOR
#	Danny Zarko. Mellanox Technologies LTD.
#
#  CREATION DATE
#	00:09:30 - 04/05/2008
#
#  MODIFICATION HISTORY
#  Initial Revision.
#
#  NOTES
#
#******
######################################################################

set debug 1
proc Start_MlxI2C {} {
    global DZ env 
    ## Set the InfoArg Array
    Set_InfoArgv

    ## Set the Tools Flags Array
    Set_ToolsFlags

    ### parsing command line arguments
    Parse_Argv 
}

proc Finish_MlxI2C {} {
    global DZ
    catch { { close $DZ(logFileID) }}
    exit 0
}

proc Initialize_MlxI2C {} {
    global DZ argv argv0 InfoArgv 

    ### Set general vars
    #set DZ(var:default.fw.dir) /mswg/projects/fw-25408/last_revision

    set DZ(var:version.num) 1.0.0
    set DZ(var:tool.name) [file rootname [file tail $argv0]]
    set DZ(var:start.clock.seconds) [clock seconds]
    set DZ(var:list.files.extention) "log list query"

    set DZ(var:cmd.i2c) "/usr/bin/i2c"
    set DZ(var:cmd.mdevices) "/usr/mst/src/mdevices"
    set DZ(var:cmd.i2c_slaves) "/usr/bin/isw"

    set DZ(var:maximal.integer) 0xffff
    set DZ(var:minimal.integer) 0

    fconfigure stdout -buffering none

    source [file join [file dirname [info script]] mlx_i2c_if.tcl]
    source [file join [file dirname [info script]] External_apps.tcl]
    
    ### configuration of constants

    ### Reseting DZ vars
    set DZ(conf:list.mst.device) [Get_ListOfDevices]
    
    ### Reseting boolean vars
    #set DZ(bool:detect.bad.links) 0
}

### General procs
# proc Get_OffsetElement _list _firstElement {offset 1}
proc Get_OffsetElement {_list _firstElement {offset 1}} {
    set place [lsearch $_list $_firstElement]
    if {$place == -1} { return ""}
    incr place $offset
    if {($place < 0) || ($place >= [llength $_list])} {return ""}
    return [lindex $_list $place]
}

proc DupChar {_char _itt} {
    set tmp ""
    for {set i 0} {$i < $_itt} {incr i} { append tmp $_char }
    return $tmp
}

proc Bool_WordInList { word list } {
    return [expr [lsearch -exact $list $word] >= 0 ] 
}

proc Hex2Bin { hex_num {bit_length 0}} {
    set bin_list ""
    array set hexbit2bin { 
	0 0000  1 0001  2 0010  3 0011
	4 0100  5 0101  6 0110  7 0111
	8 1000  9 1001  a 1010  b 1011
	c 1100  d 1101  e 1110  f 1111
    }
    regsub {^0x} [string tolower $hex_num] {} hex_num
    set bin ""
    foreach hbit [split $hex_num ""] { 
	append bin $hexbit2bin($hbit)
    }
    if {$bit_length} {
        set zero_line [string repeat 0 $bit_length]
        set bin ${zero_line}${bin}
        incr bit_length -1 
        set bin [string range $bin end-$bit_length end]
    }
    return $bin
}

proc Bin2Hex { bin_list } {
    set bin2HexList "0000 0001 0010 0011 0100 0101 0110 0111 1000 1001 1010 1011 1100 1101 1110 1111"
    set res ""
    if {[string length $bin_list] == 0} {
        return -1
    }
    set addThis [expr [string length $bin_list] % 4]
    if {$addThis} {
        set bin_list [string repeat 0 [expr 4 - $addThis]]$bin_list
    }
    for {set i 0} {$i < [string length $bin_list]} {incr i 4} {
        set currentByte [string range $bin_list end-[expr $i + 3] end-$i]
        set res [format %x [lsearch $bin2HexList $currentByte]]$res
    }
    set res 0x$res
    return $res
}

proc wait { {mode 0} args } {
    global DZ

    set desc [info exists DZ(argv:verbose)]
    if {[lsearch $args -desc] != -1} {
        set desc 1
    }

    if {[string is integer $mode]} {
        if {$desc} {
            puts "waiting [expr $mode/1000.0] Seconds"
        }
        after $mode
	return 0
    } elseif {$mode == "pause"} {
	puts "-P- Press Any Key To Continue"
	gets stdin x
    } elseif {$mode == "MSG_PAUSE"} {
	puts [Get_OffsetElement $args -msg]
	gets stdin x
    } else {
	scan [split $mode ,] {%s %s %s} tmp_address condition tmp_val
	while {1} {
	    set res [Crspace_Read $tmp_address]
	    if {([expr ($res $condition $tmp_val)])} {
		return 0; 
		break;
	    } else {
		after 1
	    }
	}
    }
}

proc HexIncr {_varName {val 1}} {
    upvar $_varName x
    set x 0x[format %x [expr $x + $val]]
}

proc Dec2Hex {_val} {
    return 0x[format %x $_val]
}

proc GetFile {fileName} {
    if {[file exists ${fileName} ]} {
        set fileFD [open ${fileName} r]
        set data [read -nonewline $fileFD]
        close $fileFD
        return [split $data \n]
    }
    return ""
}

proc Clr {} {
    puts [exec clear]
}

proc SET {_varName _varValue {length 80}} {
    global debug verbose debugIndex
    if {![info exists debug]} {
        set debug 0
    }
    if {![info exists verbose]} {
        set verbose 0
    }
    if {![info exists debugIndex]} {
        set debugIndex 0
    }
    
    upvar $_varName x
    set x $_varValue
    set tmpName $_varName
    if {$debug} {
        if {$verbose} {
            set headLineLength [string length "-D- [ProcName 1]:($debugIndex) $tmpName : "]
            set length [expr $length + $headLineLength]
            set tmpName "[ProcName 1] $tmpName"
            if {[string length $tmpName] < 30} {
                set addSpace [string repeat " " [expr 30 - [string length $tmpName]]]
                append tmpName $addSpace
            }

            if {$debug == -1} {
                puts [concat "-D- ($debugIndex) $tmpName : " $x]
            } else {
                Print [concat "-D- ($debugIndex) $tmpName : " $x] $length -Allign [DupChar " " $headLineLength],1 
            }
        } else {
            if {[string length $tmpName] < 10} {
                set addSpace [string repeat " " [expr 10 - [string length $tmpName]]]
                append tmpName $addSpace
            }
            set headLineLength [string length "-D- ($debugIndex) $tmpName : "]
            set length [expr $length + $headLineLength]
            if {$debug == -1} {
                puts [concat "-D- ($debugIndex) $tmpName : " $x]
            } else {
                Print [concat "-D- ($debugIndex) $tmpName : " $x] $length -Allign [DupChar " " $headLineLength],1 
            }
        }
        incr debugIndex
    }
}

proc ProcName { args } {
    set upLevels 0
    if { $args != "" } { set upLevels $args }
    return [lindex [info level [expr [info level] -1 -$upLevels]] 0]
}

proc Abs {_expr} {
    set a [expr $_expr]
    if {$a < 0} {
        return [expr $a * -1]
    } else {
        return $a
    }
}

proc InitializeOutputFile {_fileName} {
    global DZ

    ## Initialize file - Test1.0: Check if the extension is legit
    set ext [file extension $_fileName]

    if {![info exists DZ(outfiles,$ext)]} {
	Inform "-E-outfile:not.valid" -file0 $_fileName
    }

    ## Initialize file - Test2.0: Check if the file is writable
    set outfile $DZ(outfiles,[file extension $_fileName])
    if { [file exists $outfile] && ! [file writable $outfile] } {
	Inform "-W-outfile:not.writable" -file0 $outfile -file1 $outfile.[pid]
	append DZ(outfiles,$ext) ".[pid]"
    }

    return [open $DZ(outfiles,$ext) w]
}

proc SeperateBySpaces {_list _length} {
    set res ""
    set i 0
    if {![regexp {^([0-9]+),([0-9]+)$} $_length . start end]} {
        set start $_length
        set end $_length

    }
    foreach word $_list {
        if {!$i} {
            set len [expr $start -1 - [string length $word]]
            incr i
        } else {
            set len [expr $end -1 - [string length $word]]
        }
        set res "${res}${word} [DupChar " " $len]"
    }
    set res [string trimright $res " "]
    return $res
}
