#####################################################################
#  ...
#  NAME
#       ...
#
#  COPYRIGHT
#               - Mellanox Confidential and Proprietary -
#
# Copyright (C) Jan. 2004, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
#
# Except as specifically permitted herein, no portion of the Information,
# including but not limited to object code and source code, may be reproduced,
# modified, distributed, republished or otherwise exploited in any form or by
# any means for any purpose without the prior written permission of Mellanox
# Technologies Ltd. Use of software subject to the terms and conditions
# detailed in the file "LICENSE.txt".
# End of legal section.
#
#  FUNCTION
#     ...
#
#  AUTHOR
#	Danny Zarko. Mellanox Technologies LTD.
#
#  CREATION DATE
#	00:09:30 - 04/05/2008
#
#  MODIFICATION HISTORY
#  Initial Revision.
#
#  NOTES
#
#******
######################################################################

proc Set_InfoArgv {} {
    global InfoArgv DZ
    array set InfoArgv {
        -op,name        "op.name"
        -op,param       "operation name"
        -op,desc        "Perform the given specific operation. Available ops: evb_init, evb_uninit, evb_voltage_triming"
        -op,arglen      1 

        -d,name         "mst.device"
        -d,param        "mst-device"
        -d,desc         "Provide the mst device to use"
        -d,arglen       1 
        -d,error        "-E-argv:invalid.mst.device"
        -d,long.version  "--dev"

        -c,name         "clear.route"
        -c,desc         "Clear all known I2C switches"
        -c,arglen       0 
        -c,long.version  "--clear"

        -s,name         "system.name"
        -s,param        "evb|shark|reindeer"
        -s,desc         "Specify the system type"
        -s,arglen       1
        -s,error        "-E-argv:invalid.system.name"
        -s,regexp       {^((evb)|(shark)|(reindeer))$}

        -p,name         "route.component"
        -p,param        "component-name"
        -p,desc         "Point the I2C switches to the bus of the given component name"
        -p,arglen       1
        
        -r,name         "read.address"
        -r,param        "read address/s"
        -r,desc         "Generate an I2C read command/s"
        -r,arglen       1 
        -r,error        "-E-argv:invalid.read.command"
        -r,long.version  "--read"

        -w,name         "write.address.val"
        -w,param        "address and val/s"
        -w,desc         "Generate an I2C write command/s"
        -w,arglen       1 
        -w,error        "-E-argv:invalid.write.command"
        -w,long.version  "--write"

        -l,name         "list"
        -l,desc         "Dump the reg table"
        -l,arglen       0 
        -l,long.version  "--list"

        -set_gpio,name      "set.gpio.pin"
        -set_gpio,param     "comp:pin-name=val"
        -set_gpio,desc      "Modify a GPIO pin on the provided component"
        -set_gpio,arglen     1 

        -q,name         "query"
        -q,desc         "Dump the uniqe reg table"
        -q,param        "component-name"
        -q,arglen       0 
        -q,long.version  "--query"

        -Q,name         "Query"
        -Q,param        "all|type"
        -Q,desc         "Print query for the all instances of the given type.\nDisplay the query data in a line per instance.\nIf ALL is specified, query all queryable device types"
        -Q,arglen       1 

        -a,name         "slave.address"
        -a,param        "i2c-address"
        -a,desc         "provides the required phy i2c address"
        -a,arglen       1 
        -a,regexp       "^(0x)?[0-9a-fA-F]{2}$"
        -a,error        "-E-argv:invalid.i2c.address"
        -a,long.version "--address"

        -o,name           "out.dir"	
        -o,default        "/tmp"
        -o,param          "out-dir"
        -o,desc	          "Specifies the directory where the output files will be placed"
        -o,regexp         "dir.exists.readable,writeable"
        
        -v,name           "verbose"
        -v,desc	          "Specifies the tool to run in verbose lvl"
        -v,arglen         0
        
        -h,name	          "help"
        -h,desc	          "Prints this help Information"
        -h,long.version   "--help"
        
        -V,name         "version"
        -V,desc	        "Prints the version of the tool"
        -V,long.version "--version"
    }

    set InfoArgv(-d,regexp) ^(([join $DZ(conf:list.mst.device) )|(]))$
    
    ### Edit InfoArgv regexp with given lists
    # Example 1: join a list toghter
    # set InfoArgv(-skip,regexp) ^(([join $DZ(var:list.skip.steps) )|(]))$
    # Example 2: join a list toghter with special diffrential chars 
    # set tmpReg "=(0x)?0*\[a-fA-F0-9\]+"
    # set InfoArgv(<Flag>,regexp) "^(([ join $DZ(var:list.pm.counter) $tmpReg)|( ])$tmpReg)$"

    ### Some changes from the default definitions 
    switch -exact -- $DZ(var:tool.name) { 
        "mlxi2c" {
            array set InfoArgv { 
                -l,desc "List all components on the system"
                -q,desc "Query the given name.\nFor GPIOs, print the current value and also the description for each pin"
                -q,arglen 1
            }
        }
    }
}

proc Set_ToolsFlags {} {
    global TOOLS_FLAGS
    array set TOOLS_FLAGS {
        mlxi2c       "(d) (s) . (l|c|Q|q|p|set_gpio|op) . v" 
        LM75        "(d) (a) . (r|w|q|l) . v"
        9555        "(d) (a) . (r|w|q|l) . v"
        9505        "(d) (a) . (r|w|q|l) . v"
        DS450       "(d) (a) . (r|w|q|l) . v"
        i2c_switch  "(d) (a) . (r|w|q|l) . v"
        ADM1024     "(d) (a) . (r|w|q|l) . v"
        general	    "h V"
    }
}

proc Get_ToolsFlags { tool } { 
    global TOOLS_FLAGS
    if {[info exists TOOLS_FLAGS($tool)]} {
        return $TOOLS_FLAGS($tool)        
    }
    return ""
}

proc HighPriortyFlag {} {
    global argv
    ## If help is needed ...
    if { [Bool_WordInList "-h" $argv] || [Bool_WordInList "--help" $argv] } { 
	Inform "-H-help"
    }
    ## output the application version
    if { [Bool_WordInList "-V" $argv] || [Bool_WordInList "--version" $argv] } { 
	Inform "-H-version"
    }
}

proc Set_DefaultValues {} {
    global DZ argv InfoArgv
    foreach entry [array names InfoArgv "*,default"] { 
        set flag [lindex [split $entry ,] 0]
        if {[catch { set name $InfoArgv($flag,name)}]} { continue; }
        if { $InfoArgv($entry) != "" } { 
	    set DZ(argv:$name) $InfoArgv($entry)
	}
    }
}

proc Remove_LongVersion {_argv} {
    global InfoArgv
    set long_version [array get InfoArgv *,long.version]
    for {set i 0} {$i < [llength $_argv]} {incr i} {
        set arg [lindex $_argv $i]
        set val [Get_OffsetElement $long_version $arg -1]
        if {$val == ""} {
            continue
        }
        if {![regexp {^([^,]+),long.version$} $val . new_flag]} { 
            continue;
        }
        set _argv [lreplace $_argv $i $i $new_flag]
    }
    set place [lsearch $_argv -d]
    if {$place != -1} {
        set _argv [concat [lrange $_argv $place end] [lrange $_argv 0 [expr $place-1]]]
    }
    return $_argv
}

proc Parse_Argv {} {
    global DZ argv env InfoArgv HERMON_CRSPACE ADDRESS_DESC_MAP
    ### For help/version/vars requests
    HighPriortyFlag

    ### Setting default values for command-line arguments 
    Set_DefaultValues

    ### Modify the argv the get rid of flags long.versions
    set argv [Remove_LongVersion $argv]

    ## Command line check - Test1.0: for the case when there
    # are two different operatinal modes of the tool
    # NA 

    set infoFlags [join [split [Get_ToolsFlags $DZ(var:tool.name)] "." ]]

    ## Command line check - Test2.0: Duplicated / Mandetory / Mutually exclusive flags
    foreach item $infoFlags {
	set mandatory [regsub -all {[()]} $item "" item]
	set litem "-[join [split $item |] " -"]"
	set mutexList ""
	foreach arg $argv {
	    append mutexList " " [lindex $litem [lsearch -exact $litem $arg]] 
	}
	set mutexList [lsort $mutexList]
	for { set i 0 } { $i < [llength $mutexList] } { incr i } {
	    if { [lindex $mutexList $i] == [lindex $mutexList [expr $i+1]] } {
		Inform "-E-command:flag.twice.specified" -flag [lindex $mutexList $i]
	    }
	}
	if { [llength $mutexList] > 1 } {
	    Inform "-E-command:flag.nonmutex.flags" -flags $mutexList
	}
	if { ( [llength $mutexList] == 0 ) && $mandatory } {
	    Inform "-E-command:flag.missing.mandatory.flag" -flag [list $litem]
	}
    }
    
    set argvList $argv
    regsub -all {[()|]} $infoFlags " " allLegalFlags
    set allLegalFlags "-[join $allLegalFlags { -}]"

    ## Command line check - Test3.0: Run for the rest of the argv
    while { [llength $argvList] > 0 } {
	set flag  [lindex $argvList 0]
	set value [lindex $argvList 1]

        if { ! [Bool_WordInList $flag $allLegalFlags] } {
	    Inform "-E-command:flag.unknown" -flag $flag
	}
	set argvList [lreplace $argvList 0 0]
	if {[Bool_WordInList $flag $argvList]} {
	    Inform "-E-command:flag.twice.specified" -flag $flag
	}

	set arglen 1
	catch { set arglen $InfoArgv($flag,arglen) }
	if { [llength $argvList] == 0 && ![regexp {^0} $arglen] } { 
	    Inform "-E-command:flag.without.value" -flag $flag
	}

	### Checking values validity and setting DZ(argv:$name) - the flag's value
	set regexp 1
	if { [regexp {^0} $arglen] && [regexp {^\-} $value] } {
            set value ""
	} elseif {[info exists InfoArgv($flag,regexp)]} {
            scan [split $InfoArgv($flag,regexp) .] {%s %s %s} regType sign len
            switch $regType {
                "dir" -
                "file" {
                    scan [split $InfoArgv($flag,regexp) .] {%s %s %s} regType state attributes
                    set attributes [split $attributes ,]
		    set fn $value
		    if {$state == "exists"} {
                        if { ! [file is${regType} $fn] } {
                            if {$regType == "directory"} {
                                if {[catch {file mkdir $dir} msg] } { 
    	                            Inform "-E-argv:dir.could.not.create" -flag "-o" -value $dir -errMsg [list $msg]
                                }
                            } else {
                                Inform "-E-argv:file.not.found" -flag $flag -value $fn
                            }
			}
		    }
		    foreach attribute $attributes {
			switch $attribute {
			    "readable" -
			    "writable" {
				if {![file $attribute $fn]} {
				    Inform "-E-argv:file.not.$attribute" -flag $flag -value $fn
				}
			    }
			    "not_readable" -
			    "not_writable" {
				set attribute [lindex [split $attribute _] end]
				if {[file $attribute $fn]} {
				    Inform "-E-argv:file.not.$attribute" -flag $flag -value $fn
				}
			    }
			}
		    }
		}
                "integer" {
		    ### Turn values like 010 into 10 (instead of the tcl value - 8)
		    set int "(0x)?0*(\[a-fA-F0-9\]+)"
		    set valuesList [split $value ","]
		    ### Checking if length of integers list is OK
		    foreach condition [split $len &] {
			if { ! [expr [llength $valuesList] $condition] } {
			    Inform "$InfoArgv($flag,error)" -flag $flag -value $value
			}
		    }
		    ### Checking that all items in list are integers
		    set formattedValue ""
		    if {[catch { set maxValue $InfoArgv($flag,maxvalue) }]} { 
			set maxValue $DZ(var:maximal.integer)
		    }
		    if {[catch { set minValue $InfoArgv($flag,minvalue) }]} { 
			set minValue $DZ(var:minimal.integer)
		    }
		    foreach item $valuesList {
			### Special case: I allow the command -d ""
			if { ( [llength [join $item]] == 0 ) && ( [llength $valuesList] == 1 ) } {
			    break;
			}
			if {[catch {format %x $item} err]} {
			    if {[regexp "integer value too large to represent" $err]} { 
				Inform "-E-argv:integer.too.large" -flag $flag -value $value
			    } else { 
				Inform "$InfoArgv($flag,error)" -flag $flag -value $value
			    }
			}
			if {[string length $item] > [string length $DZ(var:maximal.integer)]} {
			    Inform "-E-argv:integer.too.large" -flag $flag -value $value
			}
                        if { $sign == "pos" && $item == 0 } {
			    Inform "$InfoArgv($flag,error)" -flag $flag -value $value
			}
			if { $sign == "nonneg" && $item < 0 } {
			    Inform "$InfoArgv($flag,error)" -flag $flag -value $value
			}
			if {$item > $maxValue} {
			    Inform "-E-argv:integer.large" -flag $flag -value $value -maxValue $maxValue
			}
			if {$item < $minValue} {
			    Inform "-E-argv:integer.small" -flag $flag -value $value -minValue $minValue
			}
			lappend formattedValue [format %d $item]
		    }
		    set value [join $formattedValue ,]
		}
		default {
		    set regexp [regexp $InfoArgv($flag,regexp) "$value"]
                    if {!$regexp} {
			Inform "$InfoArgv($flag,error)" -flag $flag -value $value
		    }
		}
	    }
	}
	set name $InfoArgv($flag,name)
	if { ($arglen == "1..") || ($arglen > 1) } {
	    if { [set next_flagIndx [lsearch -regexp $argvList {^\-}]] == -1 } { 
		set next_flagIndx [llength $argvList]
	    }
	    append DZ(argv:$name) " " [lindex $argvList 0]
	    set argvList [lreplace $argvList 0 0]
	    if {$next_flagIndx > 1} {
		set argvList [concat $flag $argvList]
	    }
	} elseif { $arglen == "0" || ( ( $arglen == "0..1" ) && ( ( $regexp==0 ) || ( $value=="" ) ) ) } {
	    set DZ(argv:$name) 1
	} elseif { $regexp } {
	    set DZ(argv:$name) "$value"
	    set argvList [lreplace $argvList 0 0]
	} else {
	    Inform "$InfoArgv($flag,error)" -flag $flag -value $value
	}
    }

    ## Command line check - Test4.0: directories and files
    if {[info exists DZ(argv:out.dir)]} { 
	foreach extention $DZ(var:list.files.extention) {
	    set DZ(outfiles,.${extention}) [file join $DZ(argv:out.dir) $DZ(var:tool.name).${extention}]
	}
    }
    
    ## Command line check - Test5.0: log file
    if {[catch {set DZ(logFileID) [open $DZ(outfiles,.log) w]} errMsg]} {
	Inform "-E-loading:cannot.open.file" $DZ(outfiles,.log) -fn $DZ(outfiles,.log) -errMsg $errMsg
    }

    # Command line check - Test6.0: I2C slaves
    set valid_slaves [Get_ListOfSlaves $DZ(argv:mst.device)]
    if {[info exists DZ(argv:slave.address)]} {
        if {[lsearch $valid_slaves $DZ(argv:slave.address)] == -1} {
            Inform "-E-argv:slave.address.illegal" -flag -s -value $DZ(argv:slave.address) -value_list $valid_slaves
        }
    }

    if {[string compare [package provide Tcl] 8.4] < 0} {
	Inform "-E-loading:cannot.use.current.tcl.package" -version [package provide Tcl]
    }
    
    return
}

proc Inform { msgCode args } {
    global DZ InfoArgv argv env
    regexp {^(\-[A-Z]\-)?([^:]+).*$} $msgCode . msgType msgSource

    if { $msgType == "-V-" } { 
	if { ![info exists DZ(argv:verbose)] } { 
	    return 
	}
    }
    ### Setting Error Codes
    set DZ(status:legal)    		0
    set DZ(status:high.priorty)		0
    set DZ(status:illegal.command)	1
    set DZ(status:illegal.flag.value)	2
    set DZ(status:mst)                  3
    set DZ(status:conf)                 4
    
    ##################################################
    ### When general tool's info is requested (help page, version num etc.)
    if { $msgType == "-H-" } { 
	switch -exact -- $msgCode { 
	    "-H-help" {
		Show_HelpPage
	    } "-H-version" { 
		append msgText "-I- $DZ(var:tool.name) version $DZ(var:version.num)"
            }
        }
	catch { Print $msgText }
	exit $DZ(status:high.priorty) 
    }

    set argsList $args 

    array set msgF {}
    while { [llength $argsList] > 0 } {
	if {[regsub "^-" [lindex $argsList 0] "" param]} { 
	    set value [lindex $argsList 1]
	    set msgF($param) $value ; 
	    set argsList  [lreplace $argsList 0 1]
	} else { 
	    set argsList  [lreplace $argsList 0 0]
	}
    }
    set putsFlags ""
    set llegalValMsg ""
    set bar "${msgType}[DupChar - 50]"
    if {[info exists msgF(flag)]} {
        set llegalValMsg "llegal value for $msgF(flag) option"
    }
    set msgText "$msgType "
    
    ### Decoding msgCode 
    switch -exact -- $msgCode {
        "-E-command:conflict.command.line" { 
            append msgText "Illegal command: conflict detected:%n$msgF(info_1)%n"
            append msgText "$msgF(info_2)%n"
        }
        "-E-command:flag.twice.specified" { 
            append msgText "Illegal command: Option $msgF(flag) is specified twice"
        }
        "-E-command:flag.nonmutex.flags" {
            append msgText "Illegal command: Options are mutually exclusive [join [join $msgF(flags)] ,]"
        }
        "-E-command:flag.missing.mandatory.flag" {
            if { [llength [join $msgF(flag)]] > 1 } { 
                append msgText "Illegal command: At least one option is mandatory: [join [join $msgF(flag)] ,]"
            } else { 
                append msgText "Illegal command: The following option is mandatory: $msgF(flag)"
            }
        }
        "-E-command:flag.unknown" { 
            append msgText "Illegal command: Unknown option $msgF(flag)"
        } 
        "-E-command:flag.without.value" {
            append msgText "Illegal command: Option $msgF(flag) requires an argument"
        }

        "-E-argv:slave.address.illegal" { 
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "Legal values are: $msgF(value_list)"
        }
        "-E-argv:dir.could.not.create" { 
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "Failed to create directory: $msgF(value) (for output files).%n"
            append msgText "Error message:%n\"$msgF(errMsg)\""
        }
        "-E-argv:file.not.found" { 
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "No such file"
        }
        "-E-argv:file.not.readable" { 
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "File/Dir is read protected"
        }
        "-E-argv:file.not.writable" { 
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "File/Dir is write protected"
        }
        "-E-argv:integer.too.large" {
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "Integer value too large to represent"
        }
        "-E-argv:integer.not.nonneg" {
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "(Legal value: a non negative integer number)"
        }
        "-E-argv:integer.not.pos" {
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "(Legal value: a positive integer number)"
        }
        "-E-argv:integer.large" { 
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "(Legal value: maximal legal value is $msgF(maxValue))" 
        }
        "-E-argv:integer.small" { 
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
            append msgText "(Legal value: minimal legal value is $msgF(minValue))" 
        }
        "-E-argv:no.command.line" {
            append msgText "Illegal argument: No command was given"
        }
        "-E-argv:invalid.mst.device" {
            append msgText "Illegal mst device: $msgF(value)%n"
            append msgText "[join $DZ(conf:list.mst.device) %n]"
        }
        "-E-argv:invalid.i2c.address" {
            append msgText "Illegal I2C address: $msgF(value)"
        }
        "-E-argv:invalid.read.address" {
            set noExiting 1
            append msgText "Illegal Read address: $msgF(value)"
        }
        "-E-argv:invalid.write.param" {
            #set noExiting 1
            append msgText "Illegal Write params: $msgF(value)"
        }
        "-E-argv:invalid.system.name" {
	    append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
	    append msgText "(Legal value: evb | shark)"
	}
        "-E-argv:invalid.set.gpio.params" {
	    append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
	}
	"-E-argv:invalid.opcode.param" {
            append msgText "Illegal argument: I${llegalValMsg}: $msgF(value)%n"
	}


        "-E-mlxi2c:unsupported.route" {
            append msgText "The component $msgF(name) has no supporting route"
        }
        "-E-mlxi2c:illegal.route.provided" {
            append msgText "The provided route $msgF(route) has a problem on the pair $msgF(pair)"
        }
        "-E-mlxi2c:route.failed" {
            append msgText "ROUTING FAILED route=$msgF(route) pair=$msgF(pair)"
        }
        "-E-mlxi2c:unsupported.query" {
            append msgText "The component called $msgF(name) is unsupported"
        }
        "-E-mlxi2c:can.not.source.file" {
            set noExiting 1
            append msgText "Can't source the file $msgF(fn)"
        }



        "-I-mlxi2c:Mode.header" {
            append msgText "Running in $msgF(mode) Mode"
        }
        "-I-mlxi2c:Type.header" {
            append msgText "$msgF(type) Query"
        }

        "-E-loading:cannot.open.file" {
            append msgText "The following file is write protected: $msgF(fn)%n"
            append msgText "Error message: \"$msgF(errMsg)\""
        }
        "-E-loading:cannot.use.current.tcl.package" {
            append msgText "The current Tcl version is: Tcl$msgF(version). "
            append msgText "$G(var:tool.name) requires Tcl8.4 or newer"
        }



        "-E-mst:can.not.retrive.mdevices" {
            append msgText "Can not figure out what are the existing devices%n"
            append msgText "I'm counting on you"
        }
        "-E-mst:mread.fail" {
            append msgText "mread failed with the following massage: %n$msgF(errMsg)"
        }
        "-E-mst:mwrite.fail" {
            append msgText "mwrite failed with the following massage: %n$msgF(errMsg)"
        }
        "-E-mst:mdevices.list.empty" {
            append msgText "mdevices command failed. Could not get dev list"
        }



        "-E-isw:can.not.retrive.i2c_slaves" {
            append msgText "Can not figure out what are the existing i2c address%n"
            append msgText "I'm counting on you"
        }
        "-E-isw:i2c_slaves.list.empty" {
            append msgText "isw command failed. Could not get I2C list"
        }



        "-V-exec:run.command" {
            append msgText "Running: $msgF(cmd)"
        }
        "-V-exec:command.res" {
            append msgText "Result: $msgF(res)"
        }
        "-W-exec:command.failed" {
            append msgText "Command Failed: $msgF(cmd)"
        }
        "-V-exec:command.failed" {
            append msgText "$msgF(err)"
        }


        "-E-outfile:not.valid" {
            append msgText "Output file $msgF(file0) is illegal value for $G(var:tool.name).%n"
        }
        "-W-outfile:not.writable" {
	    append msgText "Output file $msgF(file0) is write protected.%n"
	    append msgText "Writing info into $msgF(file1)"
	}


        "-I-test:test.name" {
            append msgText "Starting test: $msgF(test_name)"
        }


        "-W-gpio:writing.to.input.pin" {
            append msgText "Writing value to an input pin, $msgF(name)"
        }
        "-E-gpio:writing.illegal.pin.val" {
            append msgText "Value $msgF(val) is invalid PIN val"
        }


        "-I-dump:data" {
            set msgText "$msgF(data)"
        }

        default {
            append msgText "Unknown Inform val= $msgCode"
        }
    }

    ### Writing out the message  
    set msgText [split $msgText \n]
    if {[regexp ".header" $msgCode]} {
	set msgText [concat [list ""] [list $bar] [linsert $msgText 1 $bar]]
    }

    if { ($msgType == "-E-") || ($msgType == "-F-") } {
	if { ! [info exists noExiting] } {
	    puts ""
	    switch $msgSource {
		"discover" {
		    set Exiting "Aborting discovery"
		}
		"ibdiagpath" {
		    set Exiting "Aborting route tracing"
		}
		default {
		    set Exiting "Exiting"
		}
	    }
	    switch -exact -- $msgSource {
                "command" {
                    set showSynopsys 1
		    set exitStatus $DZ(status:illegal.command)
                }
                "argv" {
		    set exitStatus $DZ(status:illegal.flag.value)
		} 
                "conf" - 
                "loading" -
		"mst" {
		    set exitStatus $DZ(status:mst)
		}
		default {
		    set exitStatus -1
		}
	    }
	}
    }
    regsub -all {%n} "[join $msgText \n]" "\n" msgText
    ### DontTrimLine
    if {[info exists dontTrimLine]} {
        Print $msgText $putsFlags -length 160
    } else {
	Print $msgText $putsFlags
    }
    if {[info exists exitStatus]} { 
	puts "    $Exiting.\n"
	if {[info exists showSynopsys]} { 
	    Show_HelpPage -sysnopsys 
	    puts ""
	}
	catch { close $DZ(logFileID) }
	exit $exitStatus
    }
    return
}
proc Show_HelpPage { args } { 
    global DZ InfoArgv

    set helpPage(phy_std) \
	"DESCRIPTION
    phy_std enable to accses an external phy through MDIO
              
ERROR CODES
    1 - Illegal command line was provided
    2 - Failed to parse command line options
    3 - mread/mwrite failed"

    set helpPage(crius) \
	"DESCRIPTION
    crius enable to excute a full or partial tests list on 1 or 2 devices
              
ERROR CODES
    1 - Illegal command line was provided
    2 - Failed to parse command line options
    3 - mread/mwrite failed
    4 - Conf file invalid"

    set helpPage(crspace_std) \
	"DESCRIPTION
    crspace_std enable to accses the local crspace
            
ERROR CODES
    1 - Illegal command line was provided
    2 - Failed to parse command line options
    3 - mread/mwrite failed"

    set onlySynopsys [Bool_WordInList "-sysnopsys" $args]
    # NAME
    if { ! $onlySynopsys } { 
	puts "NAME\n  $DZ(var:tool.name)"
    }
    
    # SYNOPSIS
    set SYNOPSYS "SYNOPSYS\n  $DZ(var:tool.name)"
    set OPTIONS "OPTIONS"
    foreach item [Get_ToolsFlags $DZ(var:tool.name)] {
        if { $item == ";" } { 
	    append SYNOPSYS "\n\n  $DZ(var:tool.name)"
	    continue;
	}
        if { $item == "." } { 
	    append SYNOPSYS "\n    "
	    continue;
	}

	set synopsysFlags ""
	set mandatory [regsub -all {[()]} $item "" item]
	foreach flag [split $item |] { 
	    set flagNparam "-$flag"
            catch { append flagNparam "|$InfoArgv(-$flag,longflag)" }
            catch { append flagNparam " <$InfoArgv(-$flag,param)>" }
            set flagNdesc "$flagNparam:"
            if {[catch { append flagNdesc " $InfoArgv(-$flag,desc)" }]} {
                append flagNdesc " "
            }
	    catch { 
		if { ( [set defVal $InfoArgv(-$flag,default)] != "" ) \
			 && ( ! [Bool_WordInList $flag "i v"] ) } {
		    append flagNdesc " (default = $defVal)"
		}
	    }
	    if {[regexp {\(} $item]} { ; # only mandatory flags have optional flags
		catch { append flagNparam " \[$InfoArgv(-$flag,optional)\]" } 
	    }
	    lappend synopsysFlags $flagNparam
	    lappend OPTIONS $flagNdesc
	    lappend lcol [string first ":" $flagNdesc]
	}

	if { [string length "[lindex [split $SYNOPSYS \n] end] $synopsysFlags "] > 80 } { 
	    append SYNOPSYS "\n    "
	}
	if { ! $mandatory } { 
	    append SYNOPSYS " \[[join $synopsysFlags]\]"
	} elseif { [llength $synopsysFlags] == 1 } {  
	    append SYNOPSYS " [join $synopsysFlags |]"
	} else { 
	    append SYNOPSYS " \{[join $synopsysFlags |]\}"
	}
    }
    lappend OPTIONS ""
    foreach flag [Get_ToolsFlags general] {
	set flagNparam "-$flag"
        if {[regexp {^\-} $flag]} { 
	    set flagNparam "   -$flag"
	} 
	catch { append flagNparam " <$InfoArgv(-$flag,param)>" }
        catch { append flagNparam "|$InfoArgv(-$flag,longflag)" }
	set flagNdesc "$flagNparam:"
	catch { append flagNdesc " $InfoArgv(-$flag,desc)" }
	lappend OPTIONS $flagNdesc
	lappend lcol [string first ":" $flagNdesc]
    }

    puts "$SYNOPSYS"
    if {$onlySynopsys} { return }

    # OPTIONS
    regsub -all "<>" $OPTIONS "" newOPTIONS
    set colIndent [lindex [lsort -integer $lcol] end]
    set OPTIONS ""
    foreach option $newOPTIONS { 
	set colIdx [string first ":" $option]
	regsub ":" $option "[DupChar " " [expr $colIndent - $colIdx]]:" option
	lappend OPTIONS $option
    }
    lappend lcol [string first ":" $option]

    if {[info exists helpPage($DZ(var:tool.name)]} {
        set text [split $helpPage($DZ(var:tool.name)) \n] 
    } else {
        set text "Help_page_does_not_exists"
    }
    set index 0
    set read 0
    foreach line $text { 
	incr index
	if { [incr read [expr [regexp "DESCRIPTION" $line] + ( $read && ! [regexp {[^ ]} $line] ) ]] > 1 } {
	    break;
	}
    }

    puts "\n[join [lrange $text 0 [expr $index -1]] \n]"
    Print "[join $OPTIONS "\n  "]" -chars ": "
    Print "\n[join [lrange $text $index end ] \n]" -chars "-"

    return
}

proc Print { string args } {
    global DZ
    set maxLen 180
    set indent ""
    if { [llength $args] == 1 } { set args [join $args] }
    set chars       [Get_OffsetElement $args "-chars"]
    set nonewline   [Bool_WordInList $args "-nonewline"]
    if {[Bool_WordInList $args "-length"]} {
	set maxLen [Get_OffsetElement $args "-length"]
    }
    foreach line [split $string \n] {
        if { $chars != "" } {
	    if { [set idx [string first $chars $line]] >= 0 } {
                set indent "[string range $line 0 [expr $idx -1]]$chars"
                set line [string range $line [expr $idx+[string length $chars]] end]
	    }
	} else {
            regexp {^(-[A-Z]- ?| +)(.*)$} $line . indent line
	}
        set outline "$indent"
	set indent [string repeat " " [string length $indent]]
        set len80 [expr $maxLen - [string length $indent]]
	while { [string length $line] > 0 } {
	    set interval [string range $line 0 $len80]
	    if { [Bool_WordInList $args "-origlen"] || ( [string length $line] <= $len80 ) || ( [set spcIndx [string last " " $interval]] == -1 ) } { 
		append outline "$line"
		break;
	    } else { 
		append outline "[string range $line 0 [expr $spcIndx -1]]\n$indent"
		set line [string range $line [expr $spcIndx +1] end]
	    }
	}
	if { ![Bool_WordInList $args "-NoStdout"] } {
            if {$nonewline} {
		puts -nonewline $outline
	    } else {
		puts $outline
	    }
	}
        if { ![Bool_WordInList $args "-NoLogFile"] } {
            if {$nonewline} {
		catch { puts -nonewline $DZ(logFileID) $outline }
	    } else {
                catch { puts $DZ(logFileID) $outline }
	    }
        }
    }
}

